'''Read and write DNA sequences.'''
import csv
import os
from Bio import SeqIO
from Bio.Alphabet.IUPAC import ambiguous_dna
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation, ExactPosition
from Bio.SeqFeature import CompoundLocation
import coral
import coral.constants.genbank


class PrimerAnnotationError(ValueError):
    pass


class FeatureNameError(ValueError):
    pass


def read_dna(path):
    '''Read DNA from file. Uses BioPython and coerces to coral format.

    :param path: Full path to input file.
    :type path: str
    :returns: DNA sequence.
    :rtype: coral.DNA

    '''
    filename, ext = os.path.splitext(os.path.split(path)[-1])

    genbank_exts = ['.gb', '.ape']
    fasta_exts = ['.fasta', '.fa', '.fsa', '.seq']
    abi_exts = ['.abi', '.ab1']

    if any([ext == extension for extension in genbank_exts]):
        file_format = 'genbank'
    elif any([ext == extension for extension in fasta_exts]):
        file_format = 'fasta'
    elif any([ext == extension for extension in abi_exts]):
        file_format = 'abi'
    else:
        raise ValueError('File format not recognized.')

    seq = SeqIO.read(path, file_format)
    dna = coral.DNA(str(seq.seq))
    if seq.name == '.':
        dna.name = filename
    else:
        dna.name = seq.name

    # Features
    for feature in seq.features:
        try:
            dna.features.append(_seqfeature_to_coral(feature))
        except FeatureNameError:
            pass
    dna.features = sorted(dna.features, key=lambda feature: feature.start)
    # Used to use data_file_division, but it's inconsistent (not always the
    # molecule type)
    dna.circular = False
    with open(path) as f:
        first_line = f.read().split()
        for word in first_line:
            if word == 'circular':
                dna.circular = True

    return dna


def read_sequencing(directory):
    '''Read .seq and .abi/.ab1 results files from a dir.

    :param directory: Path to directory containing sequencing files.
    :type directory: str
    :returns: A list of DNA sequences.
    :rtype: coral.DNA list

    '''
    dirfiles = os.listdir(directory)
    seq_exts = ['.seq', '.abi', '.ab1']
    # Exclude files that aren't sequencing results
    seq_paths = [x for x in dirfiles if os.path.splitext(x)[1] in seq_exts]
    paths = [os.path.join(directory, x) for x in seq_paths]
    sequences = [read_dna(x).to_ss() for x in paths]

    return sequences


def write_dna(dna, path):
    '''Write DNA to a file (genbank or fasta).

    :param dna: DNA sequence to write to file
    :type dna: coral.DNA
    :param path: file path to write. Has to be genbank or fasta file.
    :type path: str

    '''
    # Check if path filetype is valid, remember for later
    ext = os.path.splitext(path)[1]
    if ext == '.gb' or ext == '.ape':
        filetype = 'genbank'
    elif ext == '.fa' or ext == '.fasta':
        filetype = 'fasta'
    else:
        raise ValueError('Only genbank or fasta files are supported.')

    # Convert features to Biopython form
    # Information lost on conversion:
    #     specificity of feature type
    #     strandedness
    #     topology
    features = []
    for feature in dna.features:
        features.append(_coral_to_seqfeature(feature))
    # Biopython doesn't like 'None' here
    # FIXME: this is a legacy feature - remove?
    bio_id = dna.id if hasattr(dna, 'id') else ''
    # Maximum length of name is 16
    seq = SeqRecord(Seq(str(dna), alphabet=ambiguous_dna), id=bio_id,
                    name=dna.name[0:16].replace(' ', '_'), features=features,
                    description=dna.name)
    if dna.circular:
        seq.annotations['data_file_division'] = 'circular'
    else:
        seq.annotations['data_file_division'] = 'linear'

    if filetype == 'genbank':
        SeqIO.write(seq, path, 'genbank')
    elif filetype == 'fasta':
        SeqIO.write(seq, path, 'fasta')


def write_primers(primer_list, path, names=None, notes=None):
    '''Write a list of primers out to a csv file. The first three columns are
    compatible with the current IDT order form (name, sequence, notes). By
    default there are no notes, which is an optional parameter.

    :param primer_list: A list of primers.
    :type primer_list: coral.Primer list
    :param path: A path to the csv you want to write.
    :type path: str
    :param names: A list of strings to name each oligo. Must be the same length
                  as the primer_list.
    :type names: str list
    :param notes: A list of strings to provide a note for each oligo. Must be
                  the same length as the primer_list.
    :type notes: str list

    '''
    # Check for notes and names having the right length, apply them to primers
    if names is not None:
        if len(names) != len(primer_list):
            names_msg = 'Mismatch in number of notes and primers.'
            raise PrimerAnnotationError(names_msg)
        for i, name in enumerate(names):
            primer_list[i].name = name
    if notes is not None:
        if len(notes) != len(primer_list):
            notes_msg = 'Mismatch in number of notes and primers.'
            raise PrimerAnnotationError(notes_msg)
        for i, note in enumerate(notes):
            primer_list[i].note = note

    # Write to csv
    with open(path, 'w') as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(['name', 'sequence', 'notes'])
        for primer in primer_list:
            string_rep = str(primer.overhang).lower() + str(primer.anneal)
            writer.writerow([primer.name, string_rep, primer.note])


def _process_feature_type(feature_type, bio_to_coral=True):
    '''Translate genbank feature types into usable ones (currently identical).
    The feature table is derived from the official genbank spec (gbrel.txt)
    available at http://www.insdc.org/documents/feature-table

    :param feature_type: feature to convert
    :type feature_type: str
    :param bio_to_coral: from coral to Biopython (True) or the other direction
                   (False)
    :param bio_to_coral: bool
    :returns: coral version of genbank feature_type, or vice-versa.
    :rtype: str

    '''

    err_msg = 'Unrecognized feature type: {}'.format(feature_type)
    if bio_to_coral:
        try:
            name = coral.constants.genbank.TO_CORAL[feature_type]
        except KeyError:
            raise ValueError(err_msg)
    else:
        try:
            name = coral.constants.genbank.TO_BIO[feature_type]
        except KeyError:
            raise ValueError(err_msg)
    return name


def _seqfeature_to_coral(feature):
    '''Convert a Biopython SeqFeature to a coral.Feature.

    :param feature: Biopython SeqFeature
    :type feature: Bio.SeqFeature

    '''
    # Some genomic sequences don't have a label attribute
    # TODO: handle genomic cases differently than others. Some features lack
    # a label but should still be incorporated somehow.
    qualifiers = feature.qualifiers
    if 'label' in qualifiers:
        feature_name = qualifiers['label'][0]
    elif 'locus_tag' in qualifiers:
        feature_name = qualifiers['locus_tag'][0]
    else:
        raise FeatureNameError('Unrecognized feature name')
    # Features with gaps are special, require looking at subfeatures
    # Assumption: subfeatures are never more than one level deep
    if feature.location_operator == 'join':
        # Feature has gaps. Have to figure out start/stop from subfeatures,
        # calculate gap indices. A nested feature model may be required
        # eventually.
        # Reorder the sub_feature list by start location
        # Assumption: none of the subfeatures overlap so the last entry in
        # the reordered list also has the final stop point of the feature.
        # FIXME: Getting a deprecation warning about using sub_features
        # instead of feature.location being a CompoundFeatureLocation
        reordered = sorted(feature.location.parts,
                           key=lambda location: location.start)
        starts = [int(location.start) for location in reordered]
        stops = [int(location.end) for location in reordered]
        feature_start = starts.pop(0)
        feature_stop = stops.pop(-1)
        starts = [start - feature_start for start in starts]
        stops = [stop - feature_start for stop in stops]
        feature_gaps = list(zip(stops, starts))
    else:
        # Feature doesn't have gaps. Ignore subfeatures.
        feature_start = int(feature.location.start)
        feature_stop = int(feature.location.end)
        feature_gaps = []
    feature_type = _process_feature_type(feature.type)
    if feature.location.strand == -1:
        feature_strand = 1
    else:
        feature_strand = 0
    if 'gene' in qualifiers:
        gene = qualifiers['gene']
    else:
        gene = []
    if 'locus_tag' in qualifiers:
        locus_tag = qualifiers['locus_tag']
    else:
        locus_tag = []
    coral_feature = coral.Feature(feature_name, feature_start,
                                  feature_stop, feature_type,
                                  gene=gene, locus_tag=locus_tag,
                                  qualifiers=qualifiers,
                                  strand=feature_strand,
                                  gaps=feature_gaps)
    return coral_feature


def _coral_to_seqfeature(feature):
    '''Convert a coral.Feature to a Biopython SeqFeature.

    :param feature: coral Feature.
    :type feature: coral.Feature

    '''
    bio_strand = 1 if feature.strand == 1 else -1
    ftype = _process_feature_type(feature.feature_type, bio_to_coral=False)
    sublocations = []
    if feature.gaps:
        # There are gaps. Have to define location_operator and  add subfeatures
        location_operator = 'join'
        # Feature location means nothing for 'join' sequences?
        # TODO: verify
        location = FeatureLocation(ExactPosition(0), ExactPosition(1),
                                   strand=bio_strand)
        # Reconstruct start/stop indices for each subfeature
        stops, starts = zip(*feature.gaps)
        starts = [feature.start] + [start + feature.start for start in starts]
        stops = [stop + feature.start for stop in stops] + [feature.stop]
        # Build subfeatures
        for start, stop in zip(starts, stops):
            sublocation = FeatureLocation(ExactPosition(start),
                                          ExactPosition(stop),
                                          strand=bio_strand)
            sublocations.append(sublocation)
        location = CompoundLocation(sublocations, operator='join')
    else:
        # No gaps, feature is simple
        location_operator = ''
        location = FeatureLocation(ExactPosition(feature.start),
                                   ExactPosition(feature.stop),
                                   strand=bio_strand)
    qualifiers = feature.qualifiers
    qualifiers['label'] = [feature.name]
    seqfeature = SeqFeature(location, type=ftype,
                            qualifiers=qualifiers,
                            location_operator=location_operator)
    return seqfeature



'''Yeast database query functions.'''
import coral
# TODO: Use httplib instead if we only need to do one requests-style function


def fetch_yeast_locus_sequence(locus_name, flanking_size=0):
    '''Acquire a sequence from SGD http://www.yeastgenome.org.

    :param locus_name: Common name or systematic name for the locus (e.g. ACT1
                       or YFL039C).
    :type locus_name: str
    :param flanking_size: The length of flanking DNA (on each side) to return
    :type flanking_size: int

    '''
    from intermine.webservice import Service

    service = Service('http://yeastmine.yeastgenome.org/yeastmine/service')

    # Get a new query on the class (table) you will be querying:
    query = service.new_query('Gene')

    if flanking_size > 0:

        # The view specifies the output columns
        # secondaryIdentifier: the systematic name (e.g. YFL039C)
        # symbol: short name (e.g. ACT1)
        # length: sequence length
        # flankingRegions.direction: Upstream or downstream (or both) of locus
        # flankingRegions.sequence.length: length of the flanking regions
        # flankingRegions.sequence.residues: sequence of the flanking regions
        query.add_view('secondaryIdentifier', 'symbol', 'length',
                       'flankingRegions.direction',
                       'flankingRegions.sequence.length',
                       'flankingRegions.sequence.residues')

        # You can edit the constraint values below
        query.add_constraint('flankingRegions.direction', '=', 'both',
                             code='A')
        query.add_constraint('Gene', 'LOOKUP', locus_name, 'S. cerevisiae',
                             code='B')
        query.add_constraint('flankingRegions.distance', '=',
                             '{:.1f}kb'.format(flanking_size / 1000.),
                             code='C')
        # Uncomment and edit the code below to specify your own custom logic:
        query.set_logic('A and B and C')

        # TODO: What to do when there's more than one result?
        first_result = query.rows().next()
        # FIXME: Use logger module instead
        # print first_result['secondaryIdentifier']
        # print first_result['symbol'], row['length']
        # print first_result['flankingRegions.direction']
        # print first_result['flankingRegions.sequence.length']
        # print first_result['flankingRegions.sequence.residues']

        seq = coral.DNA(first_result['flankingRegions.sequence.residues'])
        # TODO: add more metadata

    elif flanking_size == 0:
        # The view specifies the output columns
        query.add_view('primaryIdentifier', 'secondaryIdentifier', 'symbol',
                       'name', 'sgdAlias', 'organism.shortName',
                       'sequence.length', 'sequence.residues', 'description',
                       'qualifier')

        query.add_constraint('status', 'IS NULL', code='D')
        query.add_constraint('status', '=', 'Active', code='C')
        query.add_constraint('qualifier', 'IS NULL', code='B')
        query.add_constraint('qualifier', '!=', 'Dubious', code='A')
        query.add_constraint('Gene', 'LOOKUP', locus_name, 'S. cerevisiae',
                             code='E')

        # Your custom constraint logic is specified with the code below:
        query.set_logic('(A or B) and (C or D) and E')

        first_result = query.rows().next()
        seq = coral.DNA(first_result['sequence.residues'])
    else:
        print 'Problem with the flanking region size....'
        seq = coral.DNA('')

    return seq


def get_yeast_sequence(chromosome, start, end, reverse_complement=False):
    '''Acquire a sequence from SGD http://www.yeastgenome.org
    :param chromosome: Yeast chromosome.
    :type chromosome: int
    :param start: A biostart.
    :type start: int
    :param end: A bioend.
    :type end: int
    :param reverse_complement: Get the reverse complement.
    :type revervse_complement: bool
    :returns: A DNA sequence.
    :rtype: coral.DNA

    '''
    import requests

    if start != end:
        if reverse_complement:
            rev_option = '-REV'
        else:
            rev_option = ''
        param_url = '&chr=' + str(chromosome) + '&beg=' + str(start) + \
                    '&end=' + str(end) + '&rev=' + rev_option
        url = 'http://www.yeastgenome.org/cgi-bin/getSeq?map=a2map' + \
            param_url

        res = requests.get(url)
        # ok... sadely, I contacted SGD and they haven;t implemented this so
        # I have to parse their yeastgenome page, but
        # it is easy between the raw sequence is between <pre> tags!

        # warning that's for the first < so we need +5!
        begin_index = res.text.index('<pre>')

        end_index = res.text.index('</pre>')
        sequence = res.text[begin_index + 5:end_index]
        sequence = sequence.replace('\n', '').replace('\r', '')
    else:
        sequence = ''

    return coral.DNA(sequence)


def get_yeast_gene_location(gene_name):
    '''Acquire the location of a gene from SGD http://www.yeastgenome.org
    :param gene_name: Name of the gene.
    :type gene_name: string
    :returns location: [int: chromosome, int:biostart, int:bioend, int:strand]
    :rtype location: list

    '''
    from intermine.webservice import Service
    service = Service('http://yeastmine.yeastgenome.org/yeastmine/service')

    # Get a new query on the class (table) you will be querying:
    query = service.new_query('Gene')

    # The view specifies the output columns
    query.add_view('primaryIdentifier', 'secondaryIdentifier', 'symbol',
                   'name', 'organism.shortName',
                   'chromosome.primaryIdentifier',
                   'chromosomeLocation.start', 'chromosomeLocation.end',
                   'chromosomeLocation.strand')

    # Uncomment and edit the line below (the default) to select a custom sort
    # order:
    # query.add_sort_order('Gene.primaryIdentifier', 'ASC')

    # You can edit the constraint values below
    query.add_constraint('organism.shortName', '=', 'S. cerevisiae',
                         code='B')
    query.add_constraint('Gene', 'LOOKUP', gene_name, code='A')

    # Uncomment and edit the code below to specify your own custom logic:
    # query.set_logic('A and B')
    chromosomes = {'chrI': 1,
                   'chrII': 2,
                   'chrIII': 3,
                   'chrIV': 4,
                   'chrV': 5,
                   'chrVI': 6,
                   'chrVII': 7,
                   'chrVIII': 8,
                   'chrIX': 9,
                   'chrX': 10,
                   'chrXI': 11,
                   'chrXII': 12,
                   'chrXIII': 13,
                   'chrXIV': 14,
                   'chrXV': 15,
                   'chrXVI': 16}
    first_result = query.rows().next()

    return [chromosomes[first_result['chromosome.primaryIdentifier']],
            first_result['chromosomeLocation.start'],
            first_result['chromosomeLocation.end'],
            int(first_result['chromosomeLocation.strand'])]


def get_gene_id(gene_name):
    '''Retrieve systematic yeast gene name from the common name.

    :param gene_name: Common name for yeast gene (e.g. ADE2).
    :type gene_name: str
    :returns: Systematic name for yeast gene (e.g. YOR128C).
    :rtype: str

    '''
    from intermine.webservice import Service

    service = Service('http://yeastmine.yeastgenome.org/yeastmine/service')

    # Get a new query on the class (table) you will be querying:
    query = service.new_query('Gene')

    # The view specifies the output columns
    query.add_view('primaryIdentifier', 'secondaryIdentifier', 'symbol',
                   'name', 'sgdAlias', 'crossReferences.identifier',
                   'crossReferences.source.name')

    # Uncomment and edit the line below (the default) to select a custom sort
    # order:
    # query.add_sort_order('Gene.primaryIdentifier', 'ASC')

    # You can edit the constraint values below
    query.add_constraint('organism.shortName', '=', 'S. cerevisiae', code='B')
    query.add_constraint('Gene', 'LOOKUP', gene_name, code='A')

    # Uncomment and edit the code below to specify your own custom logic:
    # query.set_logic('A and B')

    for row in query.rows():
        gid = row['secondaryIdentifier']
    return gid


def get_yeast_promoter_ypa(gene_name):
    '''Retrieve promoter from Yeast Promoter Atlas
    (http://ypa.csbb.ntu.edu.tw).

    :param gene_name: Common name for yeast gene.
    :type gene_name: str
    :returns: Double-stranded DNA sequence of the promoter.
    :rtype: coral.DNA

    '''
    import requests

    loc = get_yeast_gene_location(gene_name)
    gid = get_gene_id(gene_name)
    ypa_baseurl = 'http://ypa.csbb.ntu.edu.tw/do'
    params = {'act': 'download',
              'nucle': 'InVitro',
              'right': str(loc[2]),
              'left': str(loc[1]),
              'gene': str(gid),
              'chr': str(loc[0])}

    response = requests.get(ypa_baseurl, params=params)
    text = response.text
    # FASTA records are just name-sequence pairs split up by > e.g.
    # >my_dna_name
    # GACGATA
    # TODO: most of this is redundant, as we just want the 2nd record
    record_split = text.split('>')
    record_split.pop(0)
    parsed = []
    for record in record_split:
        parts = record.split('\n')
        sequence = coral.DNA(''.join(parts[1:]))
        sequence.name = parts[0]
        parsed.append(sequence)

    return parsed[1]



# -*- coding: utf-8
'''Wrapper for ViennaRNA functions.'''
# Due to inconsistency of command inputs, each command is a function
import os
import re
from subprocess import Popen, PIPE, STDOUT
import numpy as np
from coral.utils import tempdirs


# TODO: Generic structure object to return from ViennaRNA, NUPACK classes
class ViennaRNA(object):
    def __init__(self):
        self._tempdir = ''

    @tempdirs.tempdir
    def cofold(self, strand1, strand2, temp=37.0, dangles=2, nolp=False,
               nogu=False, noclosinggu=False, constraints=None,
               canonicalbponly=False, partition=-1, pfscale=None, gquad=False):
        '''Run the RNAcofold command and retrieve the result in a dictionary.

        :param strand1: Strand 1 for running RNAcofold.
        :type strand1: coral.DNA or coral.RNA
        :param strand1: Strand 2 for running RNAcofold.
        :type strand2: coral.DNA or coral.RNA
        :param temp: Temperature at which to run the calculations.
        :type temp: float
        :param dangles: How to treat dangling end energies. Set to 0 to ignore
                        dangling ends. Set to 1 to limit unpaired bases to
                        at most one dangling end (default for MFE calc). Set to
                        2 (the default) to remove the limit in 1. Set to 3 to
                        allow coaxial stacking of adjacent helices in
                        .multi-loops
        :type dangles: int
        :param nolp: Produce structures without lonely pairs (isolated single
                     base pairs).
        :type nolp: bool
        :param nogu: Do not allow GU pairs.
        :type nogu: bool
        :param noclosinggu: Do not allow GU pairs at the end of helices.
        :type noclosinggu: bool
        :param constraints: Any structural constraints to use. Format is
                            defined at
                            http://www.tbi.univie.ac.at/RNA/RNAfold.1.html
        :type constraints: str
        :param canonicalbponly: Remove non-canonical base pairs from the
                                structure constraint (if applicable).
        :type canonicalbponly: bool
        :param partition: Calculates the partition function for the sequence.
        :type partition: int
        :param pfscale: Scaling factor for the partition function.
        :type pfScale: float
        :param gquad: Incorporate G-Quadruplex formation into the structure
                      prediction.
        :type gquad: bool

        :returns: Dictionary of calculated values, defaulting to values of MFE
                  ('mfe': float) and dotbracket structure ('dotbracket': str).
                  More keys are added depending on keyword arguments.
        :rtype: dict

        '''
        cmd_args = []
        cmd_kwargs = {'--temp=': str(temp)}
        cmd_kwargs['--dangles='] = dangles
        if nolp:
            cmd_args.append('--noLP')
        if nogu:
            cmd_args.append('--noGU')
        if noclosinggu:
            cmd_args.append('--noClosingGU')
        if constraints is not None:
            cmd_args.append('--constraint')
            if canonicalbponly:
                cmd_args.append('--canonicalBPonly')
        if partition:
            cmd_args.append('--partfunc')
        if pfscale is not None:
            cmd_kwargs['pfScale'] = float(pfscale)
        if gquad:
            cmd_args.append('--gquad')

        inputs = ['>strands\n{}&{}'.format(str(strand1), str(strand2))]
        if constraints is not None:
            inputs.append(constraints)

        rnafold_output = self._run('RNAcofold', inputs, cmd_args, cmd_kwargs)

        # Process the output
        output = {}
        lines = rnafold_output.splitlines()
        # Line 1 is the name of the sequence input, line 2 is the sequence
        lines.pop(0)
        lines.pop(0)
        # Line 3 is the dotbracket + mfe for strand1
        line3 = lines.pop(0)
        output['dotbracket'] = self._lparse(line3, '^(.*) \(')
        output['mfe'] = float(self._lparse(line3, ' \((.*)\)$'))
        # Optional outputs
        if partition:
            # Line 4 is 'a coarse representation of the pair probabilities' and
            # the ensemble free energy
            line4 = lines.pop(0)
            output['coarse'] = self._lparse(line4, '^(.*) \[')
            output['ensemble'] = float(self._lparse(line4, ' \[(.*)\]$'))
            # Line 5 is the centroid structure, its free energy, and distance
            # to the ensemble
            line5 = lines.pop(0)
            'ensemble (.*),'
            output['frequency'] = float(self._lparse(line5, 'ensemble (.*),'))
            output['deltaG'] = float(self._lparse(line5, 'binding=(.*)$'))
            # Parse the postscript file (the only place the probability matrix
            # is)
            with open(os.path.join(self._tempdir, 'strands_dp.ps')) as f:
                pattern = 'start of base pair probability data\n(.*)\nshowpage'
                dotplot_file = f.read()
                dotplot_data = re.search(pattern, dotplot_file,
                                         flags=re.DOTALL).group(1).split('\n')
                # Dimension of the dotplot - compares seq1, seq2 to self and
                # to each other (concatenation of seq1 and seq2 = axis)
                dim = len(strand1) + len(strand2)
                ensemble_probs = np.zeros((dim, dim))
                optimal_probs = np.zeros((dim, dim))

                for point in dotplot_data:
                    point_split = point.split(' ')
                    # Use zero indexing
                    i = int(point_split[0]) - 1
                    j = int(point_split[1]) - 1
                    sqprob = float(point_split[2])
                    probtype = point_split[3]
                    if probtype == 'ubox':
                        ensemble_probs[i][j] = sqprob**2
                    else:
                        optimal_probs[i][j] = sqprob**2
                output['ensemble_matrix'] = ensemble_probs
                output['optimal_matrix'] = optimal_probs

        return output

    @tempdirs.tempdir
    def fold(self, strand, temp=37.0, dangles=2, nolp=False, nogu=False,
             noclosinggu=False, constraints=None, canonicalbponly=False,
             partition=False, pfscale=None, imfeelinglucky=False, gquad=False):
        '''Run the RNAfold command and retrieve the result in a dictionary.

        :param strand: The DNA or RNA sequence on which to run RNAfold.
        :type strand: coral.DNA or coral.RNA
        :param temp: Temperature at which to run the calculations.
        :type temp: float
        :param dangles: How to treat dangling end energies. Set to 0 to ignore
                        dangling ends. Set to 1 to limit unpaired bases to
                        at most one dangling end (default for MFE calc). Set to
                        2 (the default) to remove the limit in 1. Set to 3 to
                        allow coaxial stacking of adjacent helices in
                        .multi-loops
        :type dangles: int
        :param nolp: Produce structures without lonely pairs (isolated single
                     base pairs).
        :type nolp: bool
        :param nogu: Do not allow GU pairs.
        :type nogu: bool
        :param noclosinggu: Do not allow GU pairs at the end of helices.
        :type noclosinggu: bool
        :param constraints: Any structural constraints to use. Format is
                            defined at
                            http://www.tbi.univie.ac.at/RNA/RNAfold.1.html
        :type constraints: str
        :param canonicalbponly: Remove non-canonical base pairs from the
                                structure constraint (if applicable).
        :type canonicalbponly: bool
        :param partition: Generates the partition function, generating a coarse
                          grain structure ('coarse') in the format described at
                          http://www.itc.univie.ac.at/~ivo/RNA/RNAlib/PF-Fold.h
                          tml, the ensemble free energy ('ensemble'), the
                          centroid structure ('centroid'), the free energy of
                          the centroid structure ('centroid_fe'), and its
                          distance from the ensemble ('centroid_d').
        :type partition: int
        :param pfscale: Scaling factor for the partition function.
        :type pfScale: float
        :param imfeelinglucky: Returns the one secondary structure from the
                               Boltzmann equilibrium according to its
                               probability in the ensemble.
        :type imfeelinglucky: bool
        :param gquad: Incorporate G-Quadruplex formation into the structure
                      prediction.
        :type gquad: bool

        :returns: Dictionary of calculated values, defaulting to values of MFE
                  ('mfe': float) and dotbracket structure ('dotbracket': str).
                  More keys are added depending on keyword arguments.
        :rtype: dict

        '''
        cmd_args = []
        cmd_kwargs = {'--temp=': str(temp)}
        cmd_kwargs['--dangles='] = dangles
        if nolp:
            cmd_args.append('--noLP')
        if nogu:
            cmd_args.append('--noGU')
        if noclosinggu:
            cmd_args.append('--noClosingGU')
        if constraints is not None:
            cmd_args.append('--constraint')
            if canonicalbponly:
                cmd_args.append('--canonicalBPonly')
        if partition:
            cmd_args.append('--partfunc')
        if pfscale is not None:
            cmd_kwargs['pfScale'] = float(pfscale)
        if gquad:
            cmd_args.append('--gquad')

        inputs = [str(strand)]
        if constraints is not None:
            inputs.append(constraints)

        if strand.circular:
            cmd_args.append('--circ')
        rnafold_output = self._run('RNAfold', inputs, cmd_args, cmd_kwargs)

        # Process the output
        output = {}
        lines = rnafold_output.splitlines()
        # Line 1 is the sequence as RNA
        lines.pop(0)
        # Line 2 is the dotbracket + mfe
        line2 = lines.pop(0)
        output['dotbracket'] = self._lparse(line2, '^(.*) \(')
        output['mfe'] = float(self._lparse(line2, ' \((.*)\)$'))
        # Optional outputs
        if partition:
            # Line 3 is 'a coarse representation of the pair probabilities' and
            # the ensemble free energy
            line3 = lines.pop(0)
            output['coarse'] = self._lparse(line3, '^(.*) \[')
            output['ensemble'] = float(self._lparse(line3, ' \[(.*)\]$'))
            # Line 4 is the centroid structure, its free energy, and distance
            # to the ensemble
            line4 = lines.pop(0)
            output['centroid'] = self._lparse(line4, '^(.*) \{')
            output['centroid_fe'] = float(self._lparse(line4, '^.*{(.*) d'))
            output['centroid_d'] = float(self._lparse(line4, 'd=(.*)}$'))

        return output

    def _lparse(self, line, pattern):
        '''Parses a line from STDOUT using a regex pattern.'''
        return re.search(pattern, line).group(1)

    def _run(self, command, inputs, cmd_args, cmd_kwargs):
        arguments = [command]
        arguments += cmd_args
        for flag, value in cmd_kwargs.iteritems():
            arguments.append('{} {}'.format(flag, value))

        process = Popen(arguments, stdin=PIPE, stdout=PIPE, stderr=STDOUT,
                        cwd=self._tempdir)
        command_input = '\n'.join(inputs)
        output = process.communicate(input=command_input)[0]
        return output



# -*- coding: utf-8
'''Wrapper for NUPACK 3.0.'''
import multiprocessing
import numpy as np
import os
import re
import subprocess
import time
from coral.utils import tempdirs


class NUPACK(object):
    '''Run NUPACK functions on sequences.'''
    def __init__(self, nupack_home=None):
        '''
        :param nupack_home: NUPACK home dir. If the NUPACK commands aren't in
                            your path and the NUPACKHOME environment variable
                            isn't set, you can manually specify the NUPACK
                            directory here (the directory that contains bin/).
        :type nupack_home: str

        '''
        # Figure out where the NUPACK executables are
        if nupack_home is not None:
            self._nupack_home = nupack_home
        else:
            try:
                self._nupack_home = os.environ['NUPACKHOME']
            except KeyError:
                pfunc_paths = []
                for path in os.environ['PATH'].split(os.pathsep):
                    test = os.path.join(path, 'pfunc')
                    if os.path.isfile(test) and os.access(test, os.X_OK):
                        pfunc_paths.append(test)
                if not pfunc_paths:
                    raise IOError('NUPACK commands not found - see '
                                  'documentation')

        # Initialize empty temp dir location
        self._tempdir = ''

    @tempdirs.tempdir
    def pfunc(self, strand, temp=37.0, pseudo=False, material=None,
              dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the partition function for an ordered complex of strands.
        Runs the \'pfunc\' command.

        :param strand: Strand on which to run pfunc. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the free energy of the ordered complex
                  (float) and the partition function (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command
        stdout = self._run('pfunc', cmd_args, [str(strand)]).split('\n')

        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def pfunc_multi(self, strands, permutation=None, temp=37.0, pseudo=False,
                    material=None, dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the partition function for an ordered complex of strands.
        Runs the \'pfunc\' command.

        :param strands: List of strands to use as inputs to pfunc -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the free energy of the ordered complex
                  (float) and the partition function (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        stdout = self._run('pfunc', cmd_args, lines).split('\n')

        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def pairs(self, strand, temp=37.0, pseudo=False, material=None,
              dangles='some', sodium=1.0, magnesium=0.0, cutoff=0.001):
        '''Compute the pair probabilities for an ordered complex of strands.
        Runs the \'pairs\' command.

        :param strand: Strand on which to run pairs. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param cutoff: Only probabilities above this cutoff appear in the
                       output.
        :type cutoff: float
        :returns: The probability matrix, where the (i, j)th entry
                  is the probability that base i is bound to base j. The matrix
                  is augmented (it's N+1 by N+1, where N is the number of bases
                  in the sequence) with an (N+1)th column containing the
                  probability that each base is unpaired.
        :rtype: numpy.array

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand)]
        self._run('pairs', cmd_args, lines)

        # Read the output from file
        with open(os.path.join(self._tempdir, 'pairs.ppairs')) as f:
            ppairs = f.read()
        data = re.search('\n\n\d*\n(.*)', ppairs, flags=re.DOTALL).group(1)
        N = len(strand)
        prob_matrix = np.zeros((N, N + 1))
        data_lines = data.split('\n')
        # Remove the last line (empty)
        data_lines.pop()
        # Convert into probability matrix
        for line in data_lines:
            i, j, prob = line.split('\t')
            prob_matrix[int(i) - 1, int(j) - 1] = float(prob)

        return prob_matrix

    @tempdirs.tempdir
    def pairs_multi(self, strands, permutation=None, temp=37.0, pseudo=False,
                    material=None, dangles='some', sodium=1.0, magnesium=0.0,
                    cutoff=0.001):
        '''Compute the pair probabilities for an ordered complex of strands.
        Runs the \'pairs\' command.

        :param strands: List of strands to use as inputs to pairs -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param cutoff: Only probabilities above this cutoff appear in the
                       output.
        :type cutoff: float
        :returns: Two probability matrices: The probability matrix as in the
                  pairs method (but with a dimension equal to the sum of the
                  lengths of the sequences in the permutation), and a similar
                  probability matrix where strands of the same species are
                  considered to be indistinguishable.
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        self._run('pairs', cmd_args, lines)

        # Read the output from file
        N = sum([len(s) for s in strands])
        matrices = []
        for mat_type in ['ppairs', 'epairs']:
            with open(os.path.join(self._tempdir, 'pairs.' + mat_type)) as f:
                data = f.read()

            probs = re.search('\n\n\d*\n(.*)', data, flags=re.DOTALL).group(1)
            lines = probs.split('\n')
            # Remove the last line (empty)
            lines.pop()
            prob_matrix = np.zeros((N, N + 1))
            # Convert into probability matrix
            for line in lines:
                split = line.split('\t')
                i = int(split[0]) - 1
                j = int(split[1]) - 1
                prob = float(split[2])
                prob_matrix[i, j] = prob
            matrices.append(prob_matrix)

        return matrices

    @tempdirs.tempdir
    def mfe(self, strand, temp=37.0, pseudo=False, material=None,
            dangles='some', sodium=1.0, magnesium=0.0, degenerate=False):
        '''Compute the MFE for an ordered complex of strands. Runs the \'mfe\'
        command.

        :param strand: Strand on which to run mfe. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param degenerate: Setting to True will result in returning a list of
                           dictionaries associated with structures having the
                           same, minimal MFE value.
        :type degenerate: bool
        :returns: A dictionary with keys for 'mfe' (a float), 'dotparens'
                  (dot-parens notation of the MFE structure), and 'pairlist'
                  (a pair list notation of the MFE structure). Note that the
                  pair list will be an empty list if the MFE is unstructured.
        :rtype: dict

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)
        if degenerate:
            cmd_args.append('-degenerate')

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand)]
        self._run('mfe', cmd_args, lines)

        # Read the output from file
        with open(os.path.join(self._tempdir, 'mfe.mfe')) as f:
            structures = self._process_mfe(f.read())

        if degenerate:
            return structures
        else:
            return structures[0]

    @tempdirs.tempdir
    def mfe_multi(self, strands, permutation=None, temp=37.0, pseudo=False,
                  material=None, dangles='some', sodium=1.0, magnesium=0.0,
                  degenerate=False):
        '''Compute the MFE for an ordered complex of strands. Runs the \'mfe\'
        command.

        :param strands: Strands on which to run mfe. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param degenerate: Setting to True will result in returning a list of
                           dictionaries associated with structures having the
                           same, minimal MFE value.
        :type degenerate: bool
        :returns: A dictionary with keys for 'mfe' (a float), 'dotparens'
                  (dot-parens notation of the MFE structure), and 'pairlist'
                  (a pair list notation of the MFE structure). Note that the
                  pair list will be an empty list if the MFE is unstructured.
        :rtype: dict

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        if degenerate:
            cmd_args.append('-degenerate')

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        self._run('mfe', cmd_args, lines)

        # Read the output from file
        with open(os.path.join(self._tempdir, 'mfe.mfe')) as f:
            structures = self._process_mfe(f.read())

        if degenerate:
            return structures
        else:
            return structures[0]

    @tempdirs.tempdir
    def subopt(self, strand, gap, temp=37.0, pseudo=False, material=None,
               dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the suboptimal structures within a defined energy gap of the
        MFE. Runs the \'subopt\' command.

        :param strand: Strand on which to run subopt. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param gap: Energy gap within to restrict results, e.g. 0.1.
        :type gap: float
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A list of dictionaries of the type returned by .mfe().
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand), str(gap)]
        self._run('subopt', cmd_args, lines)

        # Read the output from file
        with open(os.path.join(self._tempdir, 'subopt.subopt')) as f:
            structures = self._process_mfe(f.read())

        return structures

    @tempdirs.tempdir
    def subopt_multi(self, strands, gap, permutation=None, temp=37.0,
                     pseudo=False, material=None, dangles='some', sodium=1.0,
                     magnesium=0.0):
        '''Compute the suboptimal structures within a defined energy gap of the
        MFE for an ordered permutation of strands. Runs the \'subopt\' command.

        :param strands: Strands on which to run subopt. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param gap: Energy gap within to restrict results, e.g. 0.1.
        :type gap: float
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A list of dictionaries of the type returned by .mfe().
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(str(gap))
        self._run('subopt', cmd_args, lines)

        # Read the output from file
        with open(os.path.join(self._tempdir, 'subopt.subopt')) as f:
            structures = self._process_mfe(f.read())

        return structures

    @tempdirs.tempdir
    def count(self, strand, pseudo=False):
        '''Enumerates the total number of secondary structures over the
        structural ensemble Ω(π). Runs the \'count\' command.

        :param strand: Strand on which to run count. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :returns: The count of the number of structures in the structural
                  ensemble.
        :rtype: int

        '''
        # Set up command flags
        if pseudo:
            cmd_args = ['-pseudo']
        else:
            cmd_args = []

        # Set up the input file and run the command
        stdout = self._run('count', cmd_args, [str(strand)]).split('\n')

        # Return the count
        return int(float(stdout[-2]))

    @tempdirs.tempdir
    def count_multi(self, strands, permutation=None, pseudo=False):
        '''Enumerates the total number of secondary structures over the
        structural ensemble Ω(π) with an ordered permutation of strands. Runs
        the \'count\' command.

        :param strands: List of strands to use as inputs to count -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :returns: Dictionary with the following key:value pairs: 'energy':
                  free energy, 'pfunc': partition function.
        :rtype: dict

        '''
        # Set up command flags
        cmd_args = ['-multi']
        if pseudo:
            cmd_args.append('-pseudo')

        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        stdout = self._run('count', cmd_args, lines).split('\n')

        return int(float(stdout[-2]))

    @tempdirs.tempdir
    def energy(self, strand, dotparens, temp=37.0, pseudo=False, material=None,
               dangles='some', sodium=1.0, magnesium=0.0):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strand: Strand on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand), dotparens]
        stdout = self._run('energy', cmd_args, lines).split('\n')

        # Return the energy
        return float(stdout[-2])

    @tempdirs.tempdir
    def energy_multi(self, strands, dotparens, permutation=None, temp=37.0,
                     pseudo=False, material=None, dangles='some', sodium=1.0,
                     magnesium=0.0):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('energy', cmd_args, lines).split('\n')

        return float(stdout[-2])

    @tempdirs.tempdir
    def prob(self, strand, dotparens, temp=37.0, pseudo=False, material=None,
             dangles='some', sodium=1.0, magnesium=0.0):
        '''Calculate the equilibrium probability of a given secondary
        structure. Runs the \'prob\' command.

        :param strand: Strand on which to run prob. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The equilibrium probability of the sequence with the
                  specified secondary structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command.
        lines = [str(strand), dotparens]
        stdout = self._run('prob', cmd_args, lines).split('\n')

        # Return the probabilities
        return float(stdout[-2])

    @tempdirs.tempdir
    def prob_multi(self, strands, dotparens, permutation=None, temp=37.0,
                   pseudo=False, material=None, dangles='some', sodium=1.0,
                   magnesium=0.0):
        '''Calculate the equilibrium probability of a given secondary
        structure of a complex of sequences in a given circular permutation.
        Runs the \'prob\' command.

        :param strands: Strands on which to run prob. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('prob', cmd_args, lines).split('\n')

        return float(stdout[-2])

    @tempdirs.tempdir
    def defect(self, strand, dotparens, temp=37.0, pseudo=False, material=None,
               dangles='some', sodium=1.0, magnesium=0.0, mfe=False):
        '''Calculate the ensemble defect for a given sequence and secondary
        structure. From the documentation, the ensemble defect is defined as
        \'the number of incorrectly paired nucleotides at equilibrium evaluated
        over the ensemble of the ordered complex.\' Runs the \'defect\'
        command.

        :param strand: Strand on which to run defect. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param mfe: Return the MFE defect (Zadeh et al., 2010) instead of the
                    ensemble defect.
        :type mfe: bool
        :returns: A length 2 list of the ensemble defect (float) and normalized
                  ensemble defect (float).
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)
        if mfe:
            cmd_args.append('-mfe')

        # Set up the input file and run the command.
        lines = [str(strand), dotparens]
        stdout = self._run('defect', cmd_args, lines).split('\n')

        # Return the defect [ensemble defect, ensemble defect]
        return [float(stdout[-3]), float(stdout[-2])]

    @tempdirs.tempdir
    def defect_multi(self, strands, dotparens, permutation=None, temp=37.0,
                     pseudo=False, material=None, dangles='some', sodium=1.0,
                     magnesium=0.0, mfe=False):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param mfe: Return the MFE defect (Zadeh et al., 2010) instead of the
                    ensemble defect.
        :type mfe: bool
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        if mfe:
            cmd_args.append('-mfe')

        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('defect', cmd_args, lines).split('\n')

        # Return the defect [ensemble defect, ensemble defect]
        return [float(stdout[-3]), float(stdout[-2])]

    # Helper methods for preparing command input files
    def _multi_lines(self, strands, permutation):
        '''Prepares lines to write to file for pfunc command input.

        :param strand: Strand input (cr.DNA or cr.RNA).
        :type strand: cr.DNA or cr.DNA
        :param permutation: Permutation (e.g. [1, 2, 3, 4]) of the type used
                            by pfunc_multi.
        :type permutation: list
        '''
        lines = []
        # Write the total number of distinct strands
        lines.append(str(len(strands)))
        # Write the distinct strands
        lines += [str(strand) for strand in strands]
        # Write the permutation
        lines.append(' '.join(str(p) for p in permutation))

        return lines

    # Helper methods for processing output files
    def _process_mfe(self, data):
        # Parse the output data
        # Find the text in between the large comment lines
        commentline = '\n% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %\n'
        # Find everything between two commentlines
        groups = data.split(commentline)
        # Everything before the comment line is notes about the command,
        # the last part is just a newline
        groups.pop(0)
        groups.pop()
        # The remainder is data
        output = []
        # Skip every other one (every 2nd match is empty lines)
        for group in groups[::2]:
            lines = group.split('\n')
            # Line 1 is the strand number (ignored)
            lines.pop(0)
            # Line 2 is the MFE
            mfe = float(lines.pop(0))
            # Line 3 is the dot-bracket structure
            dotparens = lines.pop(0)
            # If there are any more lines, they are a pair list format
            # structure
            pairlist = []
            for line in lines:
                pair = line.split('\t')
                pairlist.append([int(pair[0]) - 1, int(pair[1]) - 1])
            output.append({'mfe': mfe, 'dotparens': dotparens,
                           'pairlist': pairlist})

        return output

    # Helper methods for repetitive tasks
    def _set_material(self, strand_input, material, multi=False):
        if multi:
            if len(set([s.material for s in strand_input])) > 1:
                raise ValueError('Inputs must all be coral.DNA or all '
                                 'coral.RNA')
            if material is None:
                return strand_input[0].material
            else:
                return material
        else:
            if material is None:
                return strand_input.material
            else:
                return material

    def _prep_cmd_args(self, temp, dangles, material, pseudo, sodium,
                       magnesium, multi=False):
        cmd_args = []
        cmd_args += ['-T', temp]
        cmd_args += ['-dangles', dangles]
        cmd_args += ['-material', material]
        cmd_args += ['-sodium', sodium]
        cmd_args += ['-magnesium', magnesium]
        if pseudo:
            cmd_args.append('-pseudo')
        if multi:
            cmd_args.append('-multi')

        return cmd_args

    def _run(self, command, cmd_args, lines):
        prefix = command
        path = os.path.join(self._tempdir, '{}.in'.format(prefix))
        with open(path, 'w') as f:
            f.write('\n'.join(lines))

        arguments = [os.path.join(self._nupack_home, 'bin', command)]
        arguments += cmd_args
        arguments.append(prefix)

        arguments = [str(x) for x in arguments]
        process = subprocess.Popen(arguments, stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT, cwd=self._tempdir)
        output = process.communicate()[0]
        return output


def nupack_multi(seqs, material, cmd, arguments, report=True):
    '''Split Nupack commands over processors.

    :param inputs: List of sequences, same format as for coral.analysis.Nupack.
    :type inpus: list
    :param material: Input material: 'dna' or 'rna'.
    :type material: str
    :param cmd: Command: 'mfe', 'pairs', 'complexes', or 'concentrations'.
    :type cmd: str
    :param arguments: Arguments for the command.
    :type arguments: str
    :returns: A list of the same return value you would get from `cmd`.
    :rtype: list

    '''
    nupack_pool = multiprocessing.Pool()
    try:
        args = [{'seq': seq,
                 'cmd': cmd,
                 'material': material,
                 'arguments': arguments} for seq in seqs]
        nupack_iterator = nupack_pool.imap(run_nupack, args)
        total = len(seqs)
        msg = ' calculations complete.'
        passed = 4
        while report:
            completed = nupack_iterator._index
            if (completed == total):
                break
            else:
                if passed >= 4:
                    print '({0}/{1}) '.format(completed, total) + msg
                    passed = 0
                passed += 1
                time.sleep(1)
        multi_output = [x for x in nupack_iterator]
        nupack_pool.close()
        nupack_pool.join()
    except KeyboardInterrupt:
        nupack_pool.terminate()
        nupack_pool.close()
        raise KeyboardInterrupt

    return multi_output


def run_nupack(kwargs):
    '''Run picklable Nupack command.

    :param kwargs: keyword arguments to pass to Nupack as well as 'cmd'.
    :returns: Variable - whatever `cmd` returns.

    '''
    run = NUPACK(kwargs['seq'])
    output = getattr(run, kwargs['cmd'])(**kwargs['arguments'])
    return output



'''Evaluate windows of a sequence for in-context structure.'''
import coral.analysis


class StructureWindows(object):
    '''Evaluate windows of structure and plot the results.'''
    def __init__(self, dna):
        '''
        :param dna: DNA sequence to analyze.
        :type dna: coral.DNA

        '''
        self.template = dna
        self.walked = []
        self.core_starts = []
        self.core_ends = []
        self.scores = []

    def windows(self, window_size=60, context_len=90, step=10):
        '''Walk through the sequence of interest in windows of window_size,
        evaluate free (unbound) pair probabilities.

        :param window_size: Window size in base pairs.
        :type window_size: int
        :param context_len: The number of bases of context to use when
                            analyzing each window.
        :type context_len: int
        :param step: The number of base pairs to move for each new window.
        :type step: int

        '''
        self.walked = _context_walk(self.template, window_size, context_len,
                                    step)
        self.core_starts, self.core_ends, self.scores = zip(*self.walked)
        return self.walked

    def plot(self):
        '''Plot the results of the run method.'''
        try:
            from matplotlib import pylab
        except ImportError:
            raise ImportError('Optional dependency matplotlib not installed.')

        if self.walked:
            fig = pylab.figure()
            ax1 = fig.add_subplot(111)
            ax1.plot(self.core_starts, self.scores, 'bo-')
            pylab.xlabel('Core sequence start position (base pairs).')
            pylab.ylabel('Score - Probability of being unbound.')
            pylab.show()
        else:
            raise Exception('Run calculate() first so there\'s data to plot!')


def _context_walk(dna, window_size, context_len, step):
    '''Generate context-dependent 'non-boundedness' scores for a DNA sequence.

    :param dna: Sequence to score.
    :type dna: coral.DNA
    :param window_size: Window size in base pairs.
    :type window_size: int
    :param context_len: The number of bases of context to use when analyzing
                        each window.
    :type context_len: int
    :param step: The number of base pairs to move for each new window.
    :type step: int

    '''
    # Generate window indices
    window_start_ceiling = len(dna) - context_len - window_size
    window_starts = range(context_len - 1, window_start_ceiling, step)
    window_ends = [start + window_size for start in window_starts]

    # Generate left and right in-context subsequences
    l_starts = [step * i for i in range(len(window_starts))]
    l_seqs = [dna[start:end] for start, end in zip(l_starts, window_ends)]
    r_ends = [x + window_size + context_len for x in window_starts]
    r_seqs = [dna[start:end].reverse_complement() for start, end in
              zip(window_starts, r_ends)]

    # Combine and calculate nupack pair probabilities
    seqs = l_seqs + r_seqs
    pairs_run = coral.analysis.nupack_multi(seqs, 'dna', 'pairs', {'index': 0})
    # Focus on pair probabilities that matter - those in the window
    pairs = [run[-window_size:] for run in pairs_run]
    # Score by average pair probability
    lr_scores = [sum(pair) / len(pair) for pair in pairs]

    # Split into left-right contexts again and sum for each window
    l_scores = lr_scores[0:len(seqs) / 2]
    r_scores = lr_scores[len(seqs) / 2:]
    scores = [(l + r) / 2 for l, r in zip(l_scores, r_scores)]

    # Summarize and return window indices and score
    summary = zip(window_starts, window_ends, scores)

    return summary



'''DNA object classes.'''
import collections
import os
import shutil
import subprocess
import tempfile
import coral.analysis
import coral.reaction
import coral.seqio
from ._sequence import NucleicAcidSequence


class IPythonDisplayImportError(ImportError):
    '''Failed to import IPython display modules - display requires IPython'''


class DNA(object):
    '''DNA sequence.'''
    def __init__(self, dna, bottom=None, circular=False, ds=True,
                 features=None, run_checks=True, id=None, name=''):
        '''
        :param dna: Input sequence (DNA).
        :type dna: str
        :param bottom: Manual input of bottom-strand sequence. Enables both
                       mismatches and initializing ssDNA.
        :type bottom: str
        :param circular: The topology of the DNA - True for circular DNA, False
                         for linear DNA.
        :type circular: bool
        :param ds: The strandedness of the DNA - True for dsDNA, False for
                   ssDNA.
        :type ds: bool
        :param features: List of annotated features.
        :type features: list
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :param name: Optional name field for your DNA sequence.
        :type name: str
        :returns: coral.DNA instance.
        :rtype: coral.DNA
        :raises: ValueError if an element of `features` isn't of type
                 coral.Feature.
                 ValueError if top and bottom strands have different lengths.
                 ValueError if top and bottom strands are not complementary.

        '''
        # TODO: accept sequences in general by running str() on it
        self.material = 'dna'
        dna = dna.strip()
        self.top = NucleicAcidSequence(dna, 'dna', run_checks=run_checks)

        if not ds:
            self.bottom = NucleicAcidSequence('-' * len(self.top), 'dna',
                                              run_checks=False)
        else:
            if bottom is None:
                self.bottom = self.top.reverse_complement()
            else:
                if len(bottom) != len(self.top):
                    msg = 'Top and bottom strands are difference lengths.'
                    raise ValueError(msg)
                self.bottom = NucleicAcidSequence(bottom, 'dna',
                                                  run_checks=run_checks)
        if features is None:
            self.features = []
        else:
            self.features = features
        self.circular = circular
        self.ds = ds
        self.name = name

    def ape(self, ape_path=None):
        '''Open in ApE if `ApE` is in your command line path.'''
        # TODO: simplify - make ApE look in PATH only
        cmd = 'ApE'
        if ape_path is None:
            # Check for ApE in PATH
            ape_executables = []
            for path in os.environ['PATH'].split(os.pathsep):
                exepath = os.path.join(path, cmd)
                ape_executables.append(os.access(exepath, os.X_OK))
            if not any(ape_executables):
                raise Exception('Ape not in PATH. Use ape_path kwarg.')
        else:
            cmd = ape_path
        # Check whether ApE exists in PATH
        tmp = tempfile.mkdtemp()
        if self.name is not None and self.name:
            filename = os.path.join(tmp, '{}.ape'.format(self.name))
        else:
            filename = os.path.join(tmp, 'tmp.ape')
        coral.seqio.write_dna(self, filename)
        process = subprocess.Popen([cmd, filename])
        # Block until window is closed
        try:
            process.wait()
            shutil.rmtree(tmp)
        except KeyboardInterrupt:
            shutil.rmtree(tmp)

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely-editable copy of the current sequence.
        :rtype: coral.DNA

        '''
        # Significant performance improvements by skipping alphabet check
        features_copy = [feature.copy() for feature in self.features]
        return type(self)(str(self.top), bottom=str(self.bottom),
                          circular=self.circular, ds=self.ds,
                          features=features_copy, name=self.name,
                          run_checks=False)

    def circularize(self):
        '''Circularize linear DNA.

        :returns: A circularized version of the current sequence.
        :rtype: coral.DNA

        '''
        if self.top[-1].seq == '-' and self.bottom[0].seq == '-':
            raise ValueError('Cannot circularize - termini disconnected.')
        if self.bottom[-1].seq == '-' and self.top[0].seq == '-':
            raise ValueError('Cannot circularize - termini disconnected.')

        copy = self.copy()
        copy.circular = True
        return copy

    def display(self):
        '''Display a visualization of the sequence in an IPython notebook.'''
        try:
            from IPython.display import HTML
            import uuid
        except ImportError:
            raise IPythonDisplayImportError

        sequence_json = self.json()

        d3cdn = '//d3js.org/d3.v3.min.js'
        div_id = 'sequence_{}'.format(uuid.uuid1())

        cur_dir = os.path.abspath(os.path.dirname(__file__))
        d3_plasmid_path = os.path.join(cur_dir, 'd3-plasmid.js')
        with open(d3_plasmid_path) as f:
            d3_plasmid_js = f.read()

        html = '<div id={div_id}></div>'.format(div_id=div_id)
        js_databind = '''
        <script>
        require([\'{d3_cdn}\'], function(lib) {{
            window.data = {data};'''.format(div_id=div_id, d3_cdn=d3cdn,
                                            data=sequence_json)

        js_viz = '''
            d3sequence(window.data, \'{div_id}\')
        }});
        </script>
        '''.format(div_id=div_id)

        return HTML(html + js_databind + d3_plasmid_js + js_viz)

    def json(self):
        import json
        dna_json = {}
        dna_json['name'] = self.name
        dna_json['material'] = 'DNA'
        dna_json['circular'] = self.circular
        dna_json['sequence'] = str(self.top)
        dna_json['bottom'] = str(self.bottom)

        features_json = []
        for feature in self.features:
            feature_json = {}
            feature_json['start'] = feature.start
            feature_json['stop'] = feature.stop
            feature_json['type'] = feature.feature_type
            feature_json['name'] = feature.name
            if 'ApEinfo_fwdcolor' in feature.qualifiers:
                feature_json['color'] = feature.qualifiers['ApEinfo_fwdcolor']

            features_json.append(feature_json)
        dna_json['features'] = features_json

        return json.dumps(dna_json)

    def excise(self, feature):
        '''Removes feature from circular plasmid and linearizes. Automatically
        reorients at the base just after the feature. This operation is
        complementary to the .extract() method.

        :param feature_name: The feature to remove.
        :type feature_name: coral.Feature

        '''
        rotated = self.rotate_to_feature(feature)
        excised = rotated[feature.stop - feature.start:]

        return excised

    def extract(self, feature, remove_subfeatures=False):
        '''Extract a feature from the sequence. This operation is complementary
        to the .excise() method.

        :param feature: Feature object.
        :type feature: coral.sequence.Feature
        :param remove_subfeatures: Remove all features in the extracted
                                   sequence aside from the input feature.
        :type remove_subfeatures: bool
        :returns: A subsequence from start to stop of the feature.

        '''
        extracted = self[feature.start:feature.stop]
        # Turn gaps into Ns or Xs
        for gap in feature.gaps:
            for i in range(*gap):
                extracted[i] = self._any_char
        if remove_subfeatures:
            # Keep only the feature specified
            extracted.features = [feature]
        # Update feature locations
        # copy them
        for feature in extracted.features:
            feature.move(-feature.start)
        return extracted

    def flip(self):
        '''Flip the DNA - swap the top and bottom strands.

        :returns: Flipped DNA (bottom strand is now top strand, etc.).
        :rtype: coral.DNA

        '''
        copy = self.copy()
        copy.top, copy.bottom = copy.bottom, copy.top
        return copy

    def gc(self):
        '''Find the frequency of G and C in the current sequence.'''

        gc_n = len([base for base in self if str(base) == 'C' or
                    str(base) == 'G'])
        return float(gc_n) / len(self)

    def is_palindrome(self):
        if self.top.is_palindrome() and self.bottom.is_palindrome():
            return True
        else:
            return False

    def is_rotation(self, other):
        if len(self) != len(other):
            return False

        for i in range(len(self)):
            if self.rotate(i) == other:
                return True

        # Check reverse complement
        rc = self.reverse_complement()
        for i in range(len(self)):
            if rc.rotate(i) == other:
                return True
        return False

    def linearize(self, index=0):
        '''Linearize circular DNA at an index.

        :param index: index at which to linearize.
        :type index: int
        :returns: A linearized version of the current sequence.
        :rtype: coral.DNA
        :raises: ValueError if the input is linear DNA.

        '''
        if not self.circular:
            raise ValueError('Cannot relinearize linear DNA.')
        copy = self.copy()
        # Snip at the index
        if index:
            return copy[index:] + copy[:index]
        copy.circular = False

        return copy

    def locate(self, pattern):
        '''Find sequences matching a pattern. For a circular sequence, the
        search extends over the origin.

        :param pattern: str or NucleicAcidSequence for which to find matches.
        :type pattern: str or coral.DNA
        :returns: A list of top and bottom strand indices of matches.
        :rtype: list of lists of indices (ints)
        :raises: ValueError if the pattern is longer than either the input
                 sequence (for linear DNA) or twice as long as the input
                 sequence (for circular DNA).

        '''
        if self.circular:
            if len(pattern) >= 2 * len(self):
                raise ValueError('Search pattern longer than searchable ' +
                                 'sequence.')
            top = self.top + self.top[:len(pattern) - 1]
            bottom = self.bottom + self.bottom[:len(pattern) - 1]
        else:
            if len(pattern) > len(self):
                raise ValueError('Search pattern longer than searchable ' +
                                 'sequence.')
            top = self.top
            bottom = self.bottom

        top_matches = top.locate(pattern)
        bottom_matches = bottom.locate(pattern)

        # Adjust indices if doing circular search
        def reindex(location_indices):
            for index in location_indices:
                if index > len(self):
                    index -= len(self)
            return location_indices

        if self.circular and len(pattern) > 1:
            top_matches = reindex(top_matches)
            bottom_matches = reindex(bottom_matches)

        return [top_matches, bottom_matches]

    def mw(self):
        '''Calculate the molecular weight.

        :returns: The molecular weight of the current sequence.
        :rtype: float

        '''
        bases = str(self.top) + str(self.bottom)
        counter = collections.Counter(bases.lower())
        mw_a = counter['a'] * 313.2
        mw_t = counter['t'] * 304.2
        mw_g = counter['g'] * 289.2
        mw_c = counter['c'] * 329.2
        return mw_a + mw_t + mw_g + mw_c

    def rotate(self, n):
        '''Rotate Sequence by n bases.

        :param n: Number of bases to rotate.
        :type n: int
        :returns: The current sequence reoriented at `index`.
        :rtype: coral.DNA
        :raises: ValueError if applied to linear sequence or `index` is
                 negative.

        '''
        if not self.circular and n != 0:
            raise ValueError('Cannot rotate linear DNA')
        else:
            # Save and restored the features (rotated)
            rotated = self[-n:] + self[:-n]
            rotated.features = []
            for feature in self.features:
                feature_copy = feature.copy()
                feature_copy.move(n)
                # Adjust the start/stop if we move over the origin
                feature_copy.start = feature_copy.start % len(self)
                feature_copy.stop = feature_copy.stop % len(self)
                rotated.features.append(feature_copy)

            return rotated.circularize()

    def rotate_to(self, index):
        '''Orient DNA to index (only applies to circular DNA).

        :param index: DNA position at which to re-zero the DNA.
        :type index: int
        :returns: The current sequence reoriented at `index`.
        :rtype: coral.DNA
        :raises: ValueError if applied to linear sequence or `index` is
                 negative.

        '''
        return self.rotate(-index)

    def rotate_to_feature(self, feature):
        '''Reorient the DNA based on a feature it contains (circular DNA only).

        :param feature: A feature.
        :type feature: coral.Feature
        :returns: The current sequence reoriented at the start index of a
                  unique feature matching `featurename`.
        :rtype: coral.DNA
        :raises: ValueError if there is no feature of `featurename` or
                 more than one feature matches `featurename`.

        '''
        return self.rotate_to(feature.start)

    def reverse_complement(self):
        '''Reverse complement the DNA.

        :returns: A reverse-complemented instance of the current sequence.
        :rtype: coral.DNA

        '''
        copy = self.copy()
        # Note: if sequence is double-stranded, swapping strand is basically
        # (but not entirely) the same thing - gaps affect accuracy.
        copy.top = self.top.reverse_complement()
        copy.bottom = self.bottom.reverse_complement()

        # Adjust feature locations (invert)
        for feature in copy.features:
            # Swap strand
            if feature.strand == 1:
                feature.strand = 0
            else:
                feature.strand = 1
            # Swap start and stop
            feature.start, feature.stop = (feature.stop, feature.start)
            # Adjust start/stop to feature len
            feature.start = len(copy) - feature.start
            feature.stop = len(copy) - feature.stop

        return copy

    def select_features(self, term, by='name', fuzzy=False):
        '''Select features from the features list based on feature name,
           gene, or locus tag.
           :param term: Search term.
           :type term: str
           :param by: Feature attribute to search by. Options are 'name',
                      'gene', and 'locus_tag'.
           :type by: str
           :param fuzzy: If True, search becomes case-insensitive and will also
                         find substrings - e.g. if fuzzy search is enabled, a
                         search for 'gfp' would return a hit for a feature
                         named 'GFP_seq'.
           :type fuzzy: bool
           :returns: A list of features matched by the search.
           :rtype: list
        '''
        features = []
        if fuzzy:
            fuzzy_term = term.lower()
            for feature in self.features:
                if fuzzy_term in feature.__getattribute__(by).lower():
                    features.append(feature)
        else:
            for feature in self.features:
                if feature.__getattribute__(by) == term:
                    features.append(feature)

        return features

    def tm(self, parameters='cloning'):
        '''Find the melting temperature.

        :param parameters: The tm method to use (cloning, santalucia98,
                       breslauer86)
        :type parameters: str

        '''
        return coral.analysis.tm(self, parameters=parameters)

    def to_ss(self):
        '''Produce single stranded version of the current sequence.

        :returns: The current sequence, converted to ssDNA.
        :rtype: coral.DNA

        '''
        copy = self.copy()

        # Do nothing if already single-stranded
        if not self.ds:
            return copy

        copy.bottom = NucleicAcidSequence('-' * len(copy), 'dna',
                                          run_checks=False)
        for top, bottom in zip(copy.top, reversed(copy.bottom)):
            if top == bottom == '-':
                raise ValueError('Coercing to single-stranded would ' +
                                 'introduce a double stranded break.')
        copy.ds = False

        return copy

    def to_ds(self):
        '''Produce double stranded version of the current sequence.

        :returns: The current sequence, converted to dsDNA.
        :rtype: coral.DNA

        '''
        copy = self.copy()
        # Do nothing if already set
        if self.ds:
            return copy

        # Find strand that's all gaps (if ss this should be the case)
        reverse_seq = self.reverse_complement()
        if all([char == '-' for char in self.top]):
            copy.top = reverse_seq.bottom
        elif all([char == '-' for char in self.bottom]):
            copy.bottom = reverse_seq.top
        copy.ds = True

        return copy

    def transcribe(self):
        '''Transcribe into RNA.

        :returns: An RNA sequence transcribed from the current DNA sequence.
        :rtype: coral.RNA

        '''
        return coral.reaction.transcribe(self)

    def __add__(self, other):
        '''Add DNA together.

        :param other: instance to be added to.
        :type other: compatible sequence object (currently only DNA).
        :returns: Concatenated DNA sequence.
        :rtype: coral.DNA
        :raises: Exception if either sequence is circular.
                 Exception if concatenating a sequence with overhangs would
                 create a discontinuity.

        '''
        if self.circular or other.circular:
            raise Exception('Can only add linear DNA.')

        discontinuity = [False, False]
        if len(self) != 0 and len(other) != 0:
            # If either is empty, let things proceed anyways
            discontinuity[0] = (self.top[-1] == '-' and
                                other.bottom[-1] == '-')
            discontinuity[1] = (self.bottom[0] == '-' and
                                other.top[0] == '-')

        for_discontinuity = discontinuity[0]
        rev_discontinuity = discontinuity[1]

        if for_discontinuity or rev_discontinuity:
            msg = 'Concatenated DNA would be discontinuous.'
            raise Exception(msg)

        if self.ds or other.ds:
            ds = True
        else:
            ds = False

        tops = str(self.top + other.top)
        bottoms = str(other.bottom + self.bottom)

        self_features = [feature.copy() for feature in self.features]
        other_features = [feature.copy() for feature in other.features]
        for feature in other_features:
            feature.move(len(self))
        features = self_features + other_features

        new_instance = DNA(tops, bottom=bottoms, circular=False,
                           ds=ds, run_checks=False, features=features)

        return new_instance

    def __contains__(self, query):
        '''Defines `query in sequence` operator.

        :param query: query string or DNA sequence
        :type query: str or coral.DNA

        '''
        if query in self.top or query in self.bottom:
            return True
        else:
            return False

    def __delitem__(self, key):
        '''Delete sequence at an key.

        :param key: key to delete
        :type key: int
        :returns: The current sequence with the base at `key` removed.
        :rtype: coral.DNA

        '''
        del self.top[key]
        del self.bottom[-key]

    def __getitem__(self, key):
        '''Index and slice sequences.

        :param key: int or slice object for subsetting.
        :type key: int or slice object
        :returns: A subsequence matching the slice (`key`).
        :rtype: coral.DNA

        '''
        # Adjust features
        def in_slice(feature, key, circular=False):
            '''Classify a coral.Feature object as within a slice or not.

            :param feature: The feature to test.
            :type feature: coral.Feature
            :param key: A slice key.
            :type key: slice
            :param circular: Sets whether the parent sequence is circular.
            :type circular: bool

            '''
            if circular:
                # FIXME: once circular slicing w/ negative indices is
                # implemented, need to account for that here.
                # Decide whether a feature's location is within a slice's
                # coordinates
                if key.start is None:
                    if key.stop is None:
                        # Both are none - copying whole sequence using [:]
                        return True
                    else:
                        # The slice looks like [:stop], i.e. [0:stop]
                        if feature.stop > key.stop:
                            # feature extends beyond key stop
                            return False
                        else:
                            if feature.start > key.stop:
                                # If feature extends over origin, remove it
                                return False
                            else:
                                # Feature is between 0 and key.stop
                                return True
                else:
                    if key.stop is None:
                        # The slice looks like [start:] i.e.[start:length]
                        key_start = key.start % len(self)
                        if feature.start < key_start:
                            return False
                        else:
                            if feature.stop < key_start:
                                # Feature extends over origin - remove it
                                return False
                            else:
                                # Feature is between key.start and end of seq
                                return True
                    else:
                        # The slice looks like [key.start:key.stop]
                        if feature.start < key.start or \
                           feature.stop > key.stop:
                            return False
                        else:
                            return True
            if key.start is None:
                if key.stop is None:
                    # Copying whole sequence with [:]
                    return True
                else:
                    # Slice looks like [:key.stop]
                    if feature.stop > key.stop:
                        # Feature ends after key.stop
                        return False
                    else:
                        # Feature ends before key.stop
                        return True
            else:
                if key.stop is None:
                    # Slice looks like [key.start:]
                    if feature.start < key.start:
                        return False
                    else:
                        return True
                else:
                    # The slice looks like [key.start:key.stop]
                    if feature.start < key.start or \
                       feature.stop > key.stop:
                        return False
                    else:
                        return True

        saved_features = []
        if self.features:
            if isinstance(key, slice):
                # If a slice, remove stuff that isn't in the slice and
                # adjust feature starts/stops
                if key.step == 1 or key.step is None:
                    for feature in self.features:
                        if in_slice(feature, key, self.circular):
                            feature_copy = feature.copy()
                            if key.start:
                                feature_copy.move(-(key.start % len(self)))
                            saved_features.append(feature_copy)
                else:
                    # Don't copy any features - a non-1 step size should
                    # not leave any features instace.
                    pass
            else:
                for feature in self.features:
                    if feature.start == feature.stop == key:
                        feature_copy = feature.copy()
                        feature_copy.move(key)
                        saved_features.append(feature_copy)

        # Run __getitem__ on top and bottom sequences
        new_top = self.top.__getitem__(key)
        # TODO: improve efficiency of bottom's __getitem__ (this is a slow
        # strategy)
        new_bottom = self.bottom[::-1][key][::-1]

        copy = type(self)(str(new_top), bottom=str(new_bottom),
                          circular=False, ds=self.ds, features=saved_features,
                          name=self.name, run_checks=False)

        return copy

    def __eq__(self, other):
        '''Define equality - sequences and strandedness are the.

        :returns: Whether current sequence's (Watson and Crick), topology,
                  and strandedness are equivalent to those of another sequence.
        :rtype: bool

        '''
        tops_equal = self.top == other.top
        bottoms_equal = self.bottom == other.bottom
        stranded_equal = self.ds == other.ds
        if tops_equal and bottoms_equal and stranded_equal:
            return True
        else:
            return False

    def __hash__(self):
        # Enables the use of functions like set() - hash unique attributes
        return hash(self.top.seq + self.bottom.seq + str(self.circular))

    def __len__(self):
        return len(self.top)

    def __mul__(self, n):
        new_top = self.top * n
        new_bottom = self.bottom * n
        features_copy = [feature.copy() for feature in self.features]

        return type(self)(str(new_top), str(new_bottom),
                          circular=self.circular, ds=self.ds,
                          features=features_copy, name=self.name,
                          run_checks=False)

    def __radd__(self, other):
        if other == 0 or other is None:
            # For compatibility with sum()
            return self
        elif type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))
        return self + other

    def __repr__(self):
        '''String to print when object is called directly.'''
        top = self.top.__repr__()
        bottom = self.bottom.__repr__()[::-1]

        return '{}\n{}'.format(top, bottom)

    def __setitem__(self, index, new_value):
        '''Sets value at index to new value.

        :param index: The index at which the sequence will be modified.
        :type index: int
        :param new_value: The new value at that index
        :type new_value: str or coral.DNA
        :returns: The current sequence with the sequence at `index` replaced
                  with `new_value`.
        :rtype: coral.DNA
        :raises: ValueError if `new_value` is '-'.

        '''
        if new_value == '-':
            raise ValueError('Cannot insert gap - split sequence instead.')

        self.top[index] = new_value
        self.bottom = self.bottom[::-1][index][::-1]

    def __str__(self):
        return str(self.top)


class RestrictionSite(object):
    '''Recognition site and properties of a restriction endonuclease.'''
    def __init__(self, recognition_site, cut_site, name=None):
        '''
        :param recognition_site: Input sequence.
        :type recognition_site: coral.DNA
        :param cut_site: 0-indexed indices where DNA is nicked (top, then
                         bottom strand). For an n-sized recognition site, there
                         are n + 1 positions at which to cut.
        :type cut_site: 2-tuple.
        :param name: Identifier of this restriction site
        :type name: str
        :returns: instance of coral.RestrictionSite

        '''
        self.recognition_site = recognition_site  # require DNA object
        # cutsite is indexed to leftmost base of restriction site
        self.cut_site = cut_site  # tuple of where top/bottom strands are cut
        # optional name
        self.name = name

    def is_palindrome(self):
        '''Report whether sequence is palindromic.

        :returns: Whether the restriction site is a palindrome.
        :rtype: bool

        '''
        return self.recognition_site.is_palindrome()

    def cuts_outside(self):
        '''Report whether the enzyme cuts outside its recognition site.
        Cutting at the very end of the site returns True.

        :returns: Whether the enzyme will cut outside its recognition site.
        :rtype: bool

        '''
        for index in self.cut_site:
            if index < 0 or index > len(self.recognition_site) + 1:
                return True
        return False

    def copy(self):
        '''Return copy of the restriction site.

        :returns: A safely editable copy of the current restriction site.
        :rtype: coral.RestrictionSite

        '''
        return RestrictionSite(self.recognition_site, self.cut_site,
                               self.name)

    def __repr__(self):
        '''Represent a restriction site.'''
        site = self.recognition_site
        cut_symbols = ('|', '|')
        if not self.cuts_outside():
            top_left = str(site[0:self.cut_site[0]])
            top_right = str(site[self.cut_site[0]:])
            top_w_cut = top_left + cut_symbols[0] + top_right

            bottom_left = site[0:self.cut_site[1]].reverse_complement()
            bottom_left = str(bottom_left)[::-1]
            bottom_right = site[self.cut_site[1]:].reverse_complement()
            bottom_right = str(bottom_right)[::-1]
            bottom_w_cut = bottom_left + cut_symbols[1] + bottom_right
        else:
            return '\n'.join([site.top + ' {}'.format(self.cut_site),
                              site.bottom])

        return '\n'.join([top_w_cut, bottom_w_cut])

    def __len__(self):
        '''Defines len operator.

        :returns: Length of the recognition site.
        :rtype: int

        '''
        return len(self.recognition_site)


class Primer(object):
    '''A DNA primer - ssDNA with tm, anneal, and optional overhang.'''
    def __init__(self, anneal, tm, overhang=None, name='', note=''):
        '''
        :param anneal: Annealing sequence
        :type anneal: coral.DNA
        :param overhang: Overhang sequence
        :type overhang: coral.DNA
        :param tm: melting temperature
        :type tm: float
        :param name: Optional name of the primer. Used when writing to csv with
                     seqio.write_primers.
        :type name: str
        :param note: Optional description to associate with the primer. Used
                     when writing to csv with seqio.write_primers.
        :type note: str
        :returns: coral.Primer instance.

        '''
        self.tm = tm
        self.anneal = anneal.to_ss()
        if overhang is not None:
            self.overhang = overhang.to_ss()
        else:
            self.overhang = DNA('', ds=False)
        self.name = name
        self.note = note

    def copy(self):
        '''Generate a Primer copy.

        :returns: A safely-editable copy of the current primer.
        :rtype: coral.DNA

        '''
        return type(self)(self.anneal, self.tm, overhang=self.overhang,
                          name=self.name, note=self.note)

    def primer(self):
        '''Produce full (overhang + annealing sequence) primer sequence.

        :returns: The DNA sequence of the primer.
        :rtype: coral.DNA

        '''
        return self.overhang + self.anneal

    def __repr__(self):
        '''Representation of a primer.'''
        anneal = self.anneal.top.seq.upper()
        if self.overhang:
            overhang = self.overhang.top.seq.lower()
            return 'Primer: {} Tm: {:.2f}'.format(overhang + anneal, self.tm)
        else:
            return 'Primer: {} Tm: {:.2f}'.format(anneal, self.tm)

    def __str__(self):
        '''Coerce DNA object to string.

        :returns: A string of the full primer sequence.
        :rtype: str

        '''
        return str(self.primer())

    def __eq__(self, other):
        '''Define equality - sequences, topology, and strandedness are the
        same.

        :returns: Whether two primers have the same overhang and annealing
                  sequence.
        :rtype: bool

        '''
        anneal_equal = self.anneal == other.anneal
        overhang_equal = self.overhang == other.overhang
        if anneal_equal and overhang_equal:
            return True
        else:
            return False

    def __len__(self):
        '''Define len operator.

        :returns: The length of the full primer sequence.
        :rtype: int

        '''
        return len(self.primer())



'''Provides the Structure class for convenient structural analysis
questions.'''


from coral.analysis import ViennaRNA


class Structure(object):
    def __init__(self, mode='viennarna'):
        allowed = ['viennarna']
        if mode not in allowed:
            raise ValueError('Accepted values for mode are {}'.format(allowed))
        self.mode = mode
        self.calculator = ViennaRNA()

    def mfe(self, strand):
        if self.mode == 'viennarna':
            return self.calculator.fold(strand)['mfe']



'''Simulate building a construct by assembling oligos with PCA.'''
# FIXME: Would not catch the case where e.g. the first and second oligos
# bound each other almost perfectly, ruling out the third oligo from binding
# i.e. the assemble_oligos function does not test for conflicting overlaps


class AssemblyError(Exception):
    '''Raise exception if assembly can't complete for any reason.'''
    pass


def assemble_oligos(dna_list, reference=None):
    '''Given a list of DNA sequences, assemble into a single construct.
    :param dna_list: List of DNA sequences - they must be single-stranded.
    :type dna_list: coral.DNA list
    :param reference: Expected sequence - once assembly completed, this will
                      be used to reorient the DNA (assembly could potentially
                      occur from either side of a linear DNA construct if
                      oligos are in a random order). If this fails, an
                      AssemblyError is raised.
    :type reference: coral.DNA
    :raises: AssemblyError if it can't assemble for any reason.
    :returns: A single assembled DNA sequence
    :rtype: coral.DNA

    '''
    # FIXME: this protocol currently only supports 5' ends on the assembly
    # Find all matches for every oligo. If more than 2 per side, error.
    # Self-oligo is included in case the 3' end is self-complementary.
    # 1) Find all unique 3' binders (and non-binders).
    match_3 = [bind_unique(seq, dna_list, right=True) for i, seq in
               enumerate(dna_list)]
    # 2) Find all unique 5' binders (and non-binders).
    match_5 = [bind_unique(seq, dna_list, right=False) for i, seq in
               enumerate(dna_list)]
    # Assemble into 2-tuple
    zipped = zip(match_5, match_3)

    # 3) If none found, error out with 'oligo n has no binders'
    for i, oligo_match in enumerate(zipped):
        if not any(oligo_match):
            error = 'Oligo {} has no binding partners.'.format(i + 1)
            raise AssemblyError(error)
    # 4) There should be exactly 2 oligos that bind at 3' end but
    # not 5'.
    ends = []
    for i, (five, three) in enumerate(zipped):
        if five is None and three is not None:
            ends.append(i)
    # 5) If more than 2, error with 'too many ends'.
    if len(ends) > 2:
        raise AssemblyError('Too many (>2) end oligos found.')
    # 6) If more than 2, error with 'not enough ends'.
    if len(ends) < 2:
        raise AssemblyError('Not enough (<2) end oligos found.')
    # NOTE:If 1-4 are satisfied, unique linear assembly has been found (proof?)
    # 8) Start with first end and build iteratively
    last_index = ends[0]
    assembly = dna_list[last_index].to_ds()
    flip = True
    # This would be slightly less complicated if the sequences were tied to
    # their match info in a tuple
    # Append next region n - 1 times
    for i in range(len(dna_list) - 1):
        if flip:
            # Next oligo needs to be flipped before concatenation
            # Grab 3' match from last oligo's info
            current_index, matchlen = zipped[last_index][1]
            # Get new oligo sequence, make double-stranded for concatenation
            next_oligo = dna_list[current_index].to_ds()
            # Reverse complement for concatenation
            next_oligo = next_oligo.reverse_complement()
            # Don't reverse complement the next one
            flip = False
        else:
            # Grab 5' match from last oligo's info
            current_index, matchlen = zipped[last_index][0]
            # Get new oligo sequence, make double-stranded for concatenation
            next_oligo = dna_list[current_index].to_ds()
            # Reverse complement the next one
            flip = True
        # Trim overlap from new sequence
        next_oligo = next_oligo[(matchlen - 1):]
        # Concatenate and update last oligo's information
        assembly += next_oligo
        last_index = current_index
    if reference:
        if assembly == reference or assembly == reference.reverse_complement():
            return assembly
        else:
            raise AssemblyError('Assembly did not match reference')
    else:
        return assembly


def bind_unique(reference, query_list, min_overlap=12, right=True):
    '''(5' or 3' region on reference sequence that uniquely matches the reverse
    complement of the associated (5' or 3') region of one sequence in a list of
    query sequences.

    :param reference: Reference sequence.
    :type reference: coral.DNA
    :param query_list: List of query sequences.
    :type query_list: coral.DNA list
    :param min_overlap: Minimum overlap for a match (in bp).
    :type min_overlap: int
    :param right: Check right side of sequence (3'). False results in 5' check.
    :type right: bool
    :returns: Tuple of the indices of any matches and the size of the match in
              bp.
    :rtype: tuple of ints
    :raises: AssemblyError if more than one match is found.

    '''
    size = min_overlap
    found = []
    # Reverse complementing here provides massive speedup?
    rev_query = [seq.reverse_complement() for seq in query_list]
    while not found and not size > len(reference):
        for i, seq in enumerate(rev_query):
            if right:
                # FIXME: these getitems are the slowest part of assembly
                # Easiest speedup?
                if reference.endswith(seq[:size]):
                    found.append(i)
            else:
                if reference.startswith(seq[-size:]):
                    found.append(i)
        size += 1
    if len(found) > 1:
        raise AssemblyError('Ambiguous oligo binding')
    if not found:
        return None
    else:
        return found[0], size



# -*- coding: utf-8
'''Calculate the thermodynamic melting temperatures of nucleotide sequences.'''
from math import log, log10
from . import tm_params

# TODO: Owczarzy et al 2004 has better salt correction
# TODO: Remove sugimoto? It's missing important details (like salt correction)
# TODO: Write adjusted santalucia method. Did a fit of 20000 sequences
# comparing santalucia98 to finnzymes' modified breslauer method. As expected
# they correlate heavily with offset of -6.014 and slope of 1.335 -
# i.e. to get approximately the same result, do santalucia98, multiply by 1.335
# , and subtract 6.
# Double check those stats
# TODO: Make new hybrid method - combine santalucia unified with owczarzy
# corrections, compare to finnzymes.

# Methods that are currently verified to work using reference sequences:
#   'cloning'
#   'santalucia98'


def tm(seq, dna_conc=50, salt_conc=50, parameters='cloning'):
    '''Calculate nearest-neighbor melting temperature (Tm).

    :param seq: Sequence for which to calculate the tm.
    :type seq: coral.DNA
    :param dna_conc: DNA concentration in nM.
    :type dna_conc: float
    :param salt_conc: Salt concentration in mM.
    :type salt_conc: float
    :param parameters: Nearest-neighbor parameter set. Available options:
                       'breslauer': Breslauer86 parameters
                       'sugimoto': Sugimoto96 parameters
                       'santalucia96': SantaLucia96 parameters
                       'santalucia98': SantaLucia98 parameters
                       'cloning': breslauer without corrections
                       'cloning_sl98': santalucia98 fit to 'cloning'
    :type parameters: str
    :returns: Melting temperature (Tm) in °C.
    :rtype: float
    :raises: ValueError if parameter argument is invalid.

    '''
    if parameters == 'breslauer':
        params = tm_params.BRESLAUER
    elif parameters == 'sugimoto':
        params = tm_params.SUGIMOTO
    elif parameters == 'santalucia96':
        params = tm_params.SANTALUCIA96
    elif parameters == 'santalucia98' or parameters == 'cloning_sl98':
        params = tm_params.SANTALUCIA98
    elif parameters == 'cloning':
        params = tm_params.CLONING
    else:
        raise ValueError('Unsupported parameter set.')

    # Thermodynamic parameters
    pars = {'delta_h': params['delta_h'], 'delta_s': params['delta_s']}
    pars_error = {'delta_h': params['delta_h_err'],
                  'delta_s': params['delta_s_err']}

    # Error corrections - done first for use of reverse_complement parameters
    if parameters == 'breslauer':
        deltas = breslauer_corrections(seq, pars_error)
    elif parameters == 'sugimoto':
        deltas = breslauer_corrections(seq, pars_error)
    elif parameters == 'santalucia96':
        deltas = breslauer_corrections(seq, pars_error)
    elif parameters == 'santalucia98' or parameters == 'cloning_sl98':
        deltas = santalucia98_corrections(seq, pars_error)
    elif parameters == 'cloning':
        deltas = breslauer_corrections(seq, pars_error)
        deltas[0] += 3.4
        deltas[1] += 12.4

    # Sum up the nearest-neighbor enthalpy and entropy
    seq = str(seq).upper()

    new_delt = _pair_deltas(seq, pars)
    deltas[0] += new_delt[0]
    deltas[1] += new_delt[1]

    # Unit corrections
    salt_conc /= 1e3
    dna_conc /= 1e9
    deltas[0] *= 1e3

    # Universal gas constant (R)
    R = 1.9872

    # Supposedly this is what dnamate does, but the output doesn't match theirs
#    melt = (-deltas[0] / (-deltas[1] + R * log(dna_conc / 4.0))) +
#                          16.6 * log(salt_conc) - 273.15
#    return melt
    # Overall equation is supposedly:
    # sum{dH}/(sum{dS} + R ln(dna_conc/b)) - 273.15
    # with salt corrections for the whole term (or for santalucia98,
    # salt corrections added to the dS term.
    # So far, implementing this as described does not give results that match
    # any calculator but Biopython's

    if parameters == 'breslauer' or parameters == 'cloning':
        numerator = -deltas[0]
        # Modified dna_conc denominator
        denominator = (-deltas[1]) + R * log(dna_conc / 16.0)
        # Modified Schildkraut-Lifson equation adjustment
        salt_adjustment = 16.6 * log(salt_conc) / log(10.0)
        melt = numerator / denominator + salt_adjustment - 273.15
    elif parameters == 'santalucia98' or 'cloning_sl98':
        # TODO: dna_conc should be divided by 2.0 when dna_conc >> template
        # (like PCR)
        numerator = -deltas[0]
        # SantaLucia 98 salt correction
        salt_adjustment = 0.368 * (len(seq) - 1) * log(salt_conc)
        denominator = -deltas[1] + salt_adjustment + R * log(dna_conc / 4.0)
        melt = -deltas[0] / denominator - 273.15
    elif parameters == 'santalucia96':
        # TODO: find a way to test whether the code below matches another
        # algorithm. It appears to be correct, but need to test it.
        numerator = -deltas[0]
        denominator = -deltas[1] + R * log(dna_conc / 4.0)
        # SantaLucia 96 salt correction
        salt_adjustment = 12.5 * log10(salt_conc)
        melt = numerator / denominator + salt_adjustment - 273.15
    elif parameters == 'sugimoto':
        # TODO: the stuff below is untested and probably wrong
        numerator = -deltas[0]
        denominator = -deltas[1] + R * log(dna_conc / 4.0)
        # Sugimoto parameters were fit holding salt concentration constant
        # Salt correction can be chosen / ignored? Remove sugimoto set since
        # it's so similar to santalucia98?
        salt_correction = 16.6 * log10(salt_conc)
        melt = numerator / denominator + salt_correction - 273.15

    if parameters == 'cloning_sl98':
        # Corrections to make santalucia98 method approximate cloning method.
        # May be even better for cloning with Phusion than 'cloning' method
        melt *= 1.27329212575
        melt += -2.55585450119

    return melt


def _pair_deltas(seq, pars):
    '''Add up nearest-neighbor parameters for a given sequence.

    :param seq: DNA sequence for which to sum nearest neighbors
    :type seq: str
    :param pars: parameter set to use
    :type pars: dict
    :returns: nearest-neighbor delta_H and delta_S sums.
    :rtype: tuple of floats

    '''
    delta0 = 0
    delta1 = 0
    for i in range(len(seq) - 1):
        curchar = seq[i:i + 2]
        delta0 += pars['delta_h'][curchar]
        delta1 += pars['delta_s'][curchar]
    return delta0, delta1


def breslauer_corrections(seq, pars_error):
    '''Sum corrections for Breslauer '84 method.

    :param seq: sequence for which to calculate corrections.
    :type seq: str
    :param pars_error: dictionary of error corrections
    :type pars_error: dict
    :returns: Corrected delta_H and delta_S parameters
    :rtype: list of floats

    '''
    deltas_corr = [0, 0]
    contains_gc = 'G' in str(seq) or 'C' in str(seq)
    only_at = str(seq).count('A') + str(seq).count('T') == len(seq)
    symmetric = seq == seq.reverse_complement()
    terminal_t = str(seq)[0] == 'T' + str(seq)[-1] == 'T'

    for i, delta in enumerate(['delta_h', 'delta_s']):
        if contains_gc:
            deltas_corr[i] += pars_error[delta]['anyGC']
        if only_at:
            deltas_corr[i] += pars_error[delta]['onlyAT']
        if symmetric:
            deltas_corr[i] += pars_error[delta]['symmetry']
        if terminal_t and delta == 'delta_h':
            deltas_corr[i] += pars_error[delta]['terminalT'] * terminal_t
    return deltas_corr


def santalucia98_corrections(seq, pars_error):
    '''Sum corrections for SantaLucia '98 method (unified parameters).

    :param seq: sequence for which to calculate corrections.
    :type seq: str
    :param pars_error: dictionary of error corrections
    :type pars_error: dict
    :returns: Corrected delta_H and delta_S parameters
    :rtype: list of floats

    '''
    deltas_corr = [0, 0]
    first = str(seq)[0]
    last = str(seq)[-1]

    start_gc = first == 'G' or first == 'C'
    start_at = first == 'A' or first == 'T'
    end_gc = last == 'G' or last == 'C'
    end_at = last == 'A' or last == 'T'
    init_gc = start_gc + end_gc
    init_at = start_at + end_at

    symmetric = seq == seq.reverse_complement()

    for i, delta in enumerate(['delta_h', 'delta_s']):
        deltas_corr[i] += init_gc * pars_error[delta]['initGC']
        deltas_corr[i] += init_at * pars_error[delta]['initAT']
        if symmetric:
            deltas_corr[i] += pars_error[delta]['symmetry']
    return deltas_corr



'''Base sequence classes.'''
import re
from coral.constants.genbank import TO_CORAL
from coral.constants.molecular_bio import ALPHABETS, COMPLEMENTS


class Sequence(object):
    '''Abstract representation of single chain of molecular sequences, e.g.
       a single DNA or RNA strand or Peptide.'''
    def __init__(self, sequence, material, run_checks=True, any_char='N'):
        '''
        :param sequence: Input sequence.
        :type sequence: str
        :param material: Material type (dna, rna, peptide)
        :type material: str
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :param any_char: Character representing \'any\', e.g. N for DNA.
        :type any_char: str
        :type run_checks: bool
        :returns: coral.sequence.Sequence instance.

        '''
        if run_checks:
            self.seq = process_seq(sequence, material)
        else:
            self.seq = sequence

        self.material = material
        self.any_char = any_char

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely editable copy of the current sequence.

        '''
        # Significant performance improvements by skipping alphabet check
        return type(self)(self.seq, self.material, run_checks=False)

    def locate(self, pattern):
        '''Find sequences matching a pattern.

        :param pattern: Sequence for which to find matches.
        :type pattern: str
        :returns: Indices of pattern matches.
        :rtype: list of ints

        '''
        if len(pattern) > len(self):
            raise ValueError('Pattern too long.')
        pattern = str(pattern).upper()
        re_pattern = '(?=' + pattern + ')'
        return [index.start() for index in
                re.finditer(re_pattern, self.seq)]

    def __add__(self, other):
        '''Defines addition.

        :param other: Instance with which to sum.
        :type other: coral.sequence.Sequence
        :returns: Concatenated sequence.
        :rtype: coral.sequence.Sequence

        '''
        if type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))

        return type(self)(self.seq + other.seq, self.material,
                          run_checks=False)

    def __contains__(self, query):
        '''`x in y`.

        :param query: Query (i.e. exact pattern) sequence to find.
        :type query: str
        :returns: Whether the query is found in the current sequence.
        :rtype: bool

        '''
        query_str = str(query).upper()
        query_str = re.sub(self.any_char, '.', query_str)
        if re.search(query_str, str(self)):
            return True
        else:
            return False

    def __delitem__(self, key):
        '''Deletes sequence at location key.

        :param key: Index to delete
        :type key: int
        :returns: The current sequence with the moiety at `key` removed.
        :rtype: coral.sequence.Sequence

        '''
        sequence_list = list(self.seq)
        del sequence_list[key]
        self.seq = ''.join(sequence_list)

    def __eq__(self, other):
        '''Define == operator. True if sequences are the same.

        :param other: Other sequence.
        :type other: coral.sequence._sequence.Sequence
        :returns: Whether two sequences have the same base string (sequence).
        :rtype: bool

        '''
        if str(self) == str(other):
            return True
        else:
            return False

    def __getitem__(self, key):
        '''Indexing and slicing of sequences.

        :param key: int or slice for subsetting.
        :type key: int or slice
        :returns: Slice of the current sequence.
        :rtype: coral.sequence.Sequence

        '''
        # If empty sequence, return empty sequence
        if not self.seq or self.seq is None:
            if not isinstance(key, slice):
                raise IndexError()
            return self.copy()
        new_seq = self.seq[key]
        return type(self)(new_seq, self.material, run_checks=False)

    def __len__(self):
        '''Calculate sequence length.

        :returns: The length of the sequence.
        :rtype: int

        '''
        return len(self.seq)

    def __mul__(self, n):
        '''Concatenate copies of the sequence.

        :param n: Factor by which to multiply the sequence.
        :type n: int
        :returns: The current sequence repeated n times.
        :rtype: coral.sequence.Sequence
        :raises: TypeError if n is not an integer.

        '''
        # Input checking
        if n != int(n):
            raise TypeError('Multiplication by non-integer.')
        return sum([x for x in _decompose(self, n)])

    def __ne__(self, other):
        '''Define != operator.

        :param other: Other sequence.
        :type other: coral.sequence._sequence.Sequence
        :returns: The opposite of ==.
        :rtype: bool

        '''
        try:
            return not (self == other)
        except TypeError:
            return False

    def __radd__(self, other):
        '''Add unlike types (enables sum function).

        :param other: Object of any other type.
        :param other: coral.sequence._sequence.Sequence
        :returns: Concatenated sequence.
        :rtype: coral.sequence.Sequence

        '''
        if other == 0 or other is None:
            # For compatibility with sum()
            return self
        elif type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))
        return self + other

    def __repr__(self):
        '''String to print when object is called directly.'''
        display_bases = 40
        if len(self.seq) < 90:
            sequence = self.seq
        else:
            sequence = ''.join([self.seq[:display_bases], ' ... ',
                                self.seq[-display_bases:]])
        return str(sequence)

    def __setitem__(self, index, new_value):
        '''Sets index value to new value.

        :param index: Index to modify.
        :type index: int
        :param new_value: Value to input.
        :type new_value: str or coral.sequence._sequence.Sequence
        :returns: The current sequence with the moiety at `index` replaced
                  by `new_value`.
        :rtype: coral.sequence.Sequence

        '''
        sequence_list = list(self.seq)
        sequence_list[index] = str(type(self)(new_value, self.material))
        self.seq = ''.join(sequence_list)

    def __str__(self):
        '''Cast to string.

        :returns: A string of the current sequence
        :rtype: str

        '''
        return self.seq


class NucleicAcidSequence(Sequence):
    '''Abstract sequence container for a single nucleic acid sequence
       molecule.'''
    def reverse_complement(self):
        copy = self.copy()
        copy.seq = reverse_complement(self.seq, self.material)
        return copy

    def is_palindrome(self):
        seq_len = len(self.seq)
        if seq_len % 2 == 0:
            # Sequence has even number of bases, can test non-overlapping seqs
            wing = seq_len / 2
            l_wing = self[0: wing]
            r_wing = self[wing:]
            if l_wing == r_wing.reverse_complement():
                return True
            else:
                return False
        else:
            # Sequence has odd number of bases and cannot be a palindrome
            return False


def _decompose(string, n):
    '''Given string and multiplier n, find m**2 decomposition.

    :param string: input string
    :type string: str
    :param n: multiplier
    :type n: int
    :returns: generator that produces m**2 * string if m**2 is a factor of n
    :rtype: generator of 0 or 1

    '''
    binary = [int(x) for x in bin(n)[2:]]
    new_string = string
    counter = 1
    while counter <= len(binary):
        if binary[-counter]:
            yield new_string
        new_string += new_string
        counter += 1


class Feature(object):
    '''Represent A DNA feature - annotate and extract sequence by metadata.'''
    def __init__(self, name, start, stop, feature_type, gene='', locus_tag='',
                 qualifiers=None, strand=0, gaps=None):
        '''
        :param name: Name of the feature. Used during feature extraction.
        :type name: str
        :param start: Where the feature starts
        :type start: int
        :param stop: Where the feature stops
        :type stop: int
        :param feature_type: The type of the feature. Allowed types:
                                'coding', 'primer', 'promoter', 'terminator',
                                'rbs'
        :type name: str
        :param strand: Watson (0) or Crick (1) strand of the feature.
        :type strand: int
        :param gaps: Gap locations if the feature has gaps.
        :type gaps: list of coordinates (2-tuple/list)
        :param gene: gene attribute (Genbank standard) - gene name (e.g. galK
                     on MG1655 genome).
        :type gene: str
        :param locus_tag: locus_tag attribute (Genbank standard) - systematic
                          locus name (e.g. b0757 for galK on MG1655 genome).
        :type locus_tag: str
        :param qualifiers: Complete Genbank qualifiers key:value pairs
        :type qualifiers: dict
        :returns: coral.Feature instance.
        :raises: ValueError if `feature_type` is not in
                 coral.constants.genbank.TO_CORAL.

        '''
        self.name = name
        self.start = int(start)
        self.stop = int(stop)
        self.modified = False
        self.gene = gene
        self.locus_tag = locus_tag
        if qualifiers is None:
            self.qualifiers = {}
        else:
            self.qualifiers = qualifiers
        self.strand = strand
        if gaps is None:
            self.gaps = []
        else:
            self.gaps = gaps

        allowed_types = list(sorted(TO_CORAL.keys()))

        if feature_type in allowed_types:
            self.feature_type = feature_type
        else:
            msg1 = 'feature_type '
            msg2 = 'must be one of the following: {}'.format(allowed_types)
            raise ValueError(msg1 + msg2)

    def move(self, bases):
        '''Move the start and stop positions.

        :param bases: bases to move - can be negative
        :type bases: int

        '''
        self.start += bases
        self.stop += bases

    def copy(self):
        '''Return a copy of the Feature.

        :returns: A safely editable copy of the current feature.
        :rtype: coral.Feature

        '''
        return type(self)(self.name, self.start, self.stop, self.feature_type,
                          gene=self.gene, locus_tag=self.locus_tag,
                          qualifiers=self.qualifiers, strand=self.strand)

    def __repr__(self):
        '''Represent a feature.'''
        if self.modified:
            part1 = '(Modified) {} \'{}\' feature '.format(self.name,
                                                           self.feature_type)
        else:
            part1 = '{} \'{}\' feature '.format(self.name, self.feature_type)
        part2 = '({0} to {1}) on strand {2}'.format(self.start, self.stop,
                                                    self.strand)
        return part1 + part2

    def __eq__(self, other):
        '''Define equality.

        :returns: Whether the name and feature type are the same.
        :rtype: bool

        '''
        if self.name != other.name:
            return False
        if self.feature_type != other.feature_type:
            return False
        if self.start != other.start:
            return False
        if self.stop != other.stop:
            return False
        if self.strand != other.strand:
            return False
        if self.gaps != other.gaps:
            return False

        return True

    def __ne__(self, other):
        '''Define inequality.'''
        if not self == other:
            return True
        else:
            return False


def reverse_complement(sequence, material):
    '''Reverse complement a sequence.

    :param sequence: Sequence to reverse complement
    :type sequence: str
    :param material: dna, rna, or peptide.
    :type material: str
    '''
    code = dict(COMPLEMENTS[material])
    reverse_sequence = sequence[::-1]
    return ''.join([code[str(base)] for base in reverse_sequence])


def check_alphabet(seq, material):
    '''Verify that a given string is valid DNA, RNA, or peptide characters.

    :param seq: DNA, RNA, or peptide sequence.
    :type seq: str
    :param material: Input material - 'dna', 'rna', or 'pepide'.
    :type sequence: str
    :returns: Whether the `seq` is a valid string of `material`.
    :rtype: bool
    :raises: ValueError if `material` isn't \'dna\', \'rna\', or \'peptide\'.
             ValueError if `seq` contains invalid characters for its
             material type.

    '''
    errs = {'dna': 'DNA', 'rna': 'RNA', 'peptide': 'peptide'}
    if material == 'dna' or material == 'rna' or material == 'peptide':
        alphabet = ALPHABETS[material]
        err_msg = errs[material]
    else:
        msg = 'Input material must be \'dna\', \'rna\', or \'peptide\'.'
        raise ValueError(msg)
    # This is a bottleneck when modifying sequence - hence the run_checks
    # optional parameter in sequence objects..
    # First attempt with cython was slower. Could also try pypy.
    if re.search('[^' + alphabet + ']', seq):
        raise ValueError('Encountered a non-%s character' % err_msg)


def process_seq(seq, material):
    '''Validate and process sequence inputs.

    :param seq: input sequence
    :type seq: str
    :param material: DNA, RNA, or peptide
    :type: str
    :returns: Uppercase version of `seq` with the alphabet checked by
              check_alphabet().
    :rtype: str

    '''
    check_alphabet(seq, material)
    seq = seq.upper()
    return seq


def palindrome(seq):
    '''Test whether a sequence is palindrome.

    :param seq: Sequence to analyze (DNA or RNA).
    :type seq: coral.DNA or coral.RNA
    :returns: Whether a sequence is a palindrome.
    :rtype: bool

    '''
    seq_len = len(seq)
    if seq_len % 2 == 0:
        # Sequence has even number of bases, can test non-overlapping seqs
        wing = seq_len / 2
        l_wing = seq[0: wing]
        r_wing = seq[wing:]
        if l_wing == r_wing.reverse_complement():
            return True
        else:
            return False
    else:
        # Sequence has odd number of bases and cannot be a palindrome
        return False



'''Classes to contain and manipulate DNA, RNA, and protein sequences.'''
from ._dna import DNA
from ._dna import RestrictionSite
from ._dna import Primer
from ._peptide import Peptide
from ._rna import RNA
from ._sequence import Feature



'''Retrieve restriction enzymes from rebase.'''
import shutil
import tempfile
import urllib2
import coral


class Rebase(object):
    '''Retrieve restriction enzymes from rebase database.'''
    def __init__(self):
        self._tmpdir = None
        self.update()

    def update(self):
        '''Update definitions.'''
        # Download http://rebase.neb.com/rebase/link_withref to tmp
        self._tmpdir = tempfile.mkdtemp()
        try:
            self._rebase_file = self._tmpdir + '/rebase_file'
            print 'Downloading latest enzyme definitions'
            url = 'http://rebase.neb.com/rebase/link_withref'
            header = {'User-Agent': 'Mozilla/5.0'}
            req = urllib2.Request(url, headers=header)
            con = urllib2.urlopen(req)
            with open(self._rebase_file, 'wb') as rebase_file:
                rebase_file.write(con.read())
            # Process into self._enzyme_dict
            self._process_file()
        except urllib2.HTTPError, e:
            print 'HTTP Error: {} {}'.format(e.code, url)
            print 'Falling back on default enzyme list'
            self._enzyme_dict = coral.constants.fallback_enzymes
        except urllib2.URLError, e:
            print 'URL Error: {} {}'.format(e.reason, url)
            print 'Falling back on default enzyme list'
            self._enzyme_dict = coral.constants.fallback_enzymes
        # Process into RestrictionSite objects? (depends on speed)
        print 'Processing into RestrictionSite instances.'
        self.restriction_sites = {}
        # TODO: make sure all names are unique
        for key, (site, cuts) in self._enzyme_dict.iteritems():
            # Make a site
            try:
                r = coral.RestrictionSite(coral.DNA(site), cuts, name=key)
                # Add it to dict with name as key
                self.restriction_sites[key] = r
            except ValueError:
                # Encountered ambiguous sequence, have to ignore it until
                # coral.DNA can handle ambiguous DNA
                pass

    def get(self, name):
        '''Retrieve enzyme by name.

        :param name: Name of the restriction enzyme, e.g. EcoRV.
        :type name: str
        :returns: Restriction site matching the input name.
        :rtype: coral.RestrictionSite
        :raises: Exception when enzyme is not found in the database.

        '''
        # Looks for restriction enzyme by name
        try:
            return self.restriction_sites[name]
        except KeyError:
            raise Exception('Enzyme not found.')

    def _process_file(self):
        '''Process rebase file into dict with name and cut site information.'''
        print 'Processing file'
        with open(self._rebase_file, 'r') as f:
            raw = f.readlines()
        names = [line.strip()[3:] for line in raw if line.startswith('<1>')]
        seqs = [line.strip()[3:] for line in raw if line.startswith('<5>')]
        if len(names) != len(seqs):
            raise Exception('Found different number of enzyme names and '
                            'sequences.')
        self._enzyme_dict = {}
        for name, seq in zip(names, seqs):
            if '?' in seq:
                # Is unknown sequence, don't keep it
                pass
            elif seq.startswith('(') and seq.endswith(')'):
                # Has four+ cut sites, don't keep it
                pass
            elif '^' in seq:
                # Has reasonable internal cut sites, keep it
                top_cut = seq.index('^')
                bottom_cut = len(seq) - top_cut - 1
                site = seq.replace('^', '')
                self._enzyme_dict[name] = (site, (top_cut, bottom_cut))
            elif seq.endswith(')'):
                # Has reasonable external cut sites, keep it
                # (4-cutter also starts with '(')
                # separate site and cut locations
                site, cuts = seq.split('(')
                cuts = cuts.replace(')', '')
                top_cut, bottom_cut = [int(x) + len(site) for x in
                                       cuts.split('/')]
                self._enzyme_dict[name] = (site, (top_cut, bottom_cut))
        shutil.rmtree(self._tmpdir)



'''Utils for analysis module.'''
import coral


def sequence_type(seq):
    '''Validates a coral.sequence data type.

    :param sequence_in: input DNA sequence.
    :type sequence_in: any
    :returns: The material - 'dna', 'rna', or 'peptide'.
    :rtype: str
    :raises: ValueError

    '''
    if isinstance(seq, coral.DNA):
        material = 'dna'
    elif isinstance(seq, coral.RNA):
        material = 'rna'
    elif isinstance(seq, coral.Peptide):
        material = 'peptide'
    else:
        raise ValueError('Input was not a recognized coral.sequence object.')
    return material



# -*- coding: utf-8
'''Temporary directory helpers for scripts that call command line
applications. '''
import os
import shutil
import tempfile


def tempdir(fun):
    '''For use as a decorator of instance methods - creates a temporary dir
    named self._tempdir and then deletes it after the method runs.

    :param fun: function to decorate
    :type fun: instance method

    '''
    def wrapper(*args, **kwargs):
        self = args[0]
        if os.path.isdir(self._tempdir):
            shutil.rmtree(self._tempdir)
        self._tempdir = tempfile.mkdtemp()
        # If the method raises an exception, delete the temporary dir
        try:
            retval = fun(*args, **kwargs)
        finally:
            shutil.rmtree(self._tempdir)
        if os.path.isdir(self._tempdir):
            shutil.rmtree(self._tempdir)
        return retval
    return wrapper



'''Coral: code-ify your synthetic DNA design workflow.

Documentation available at http://coral.readthedocs.org.

Coral is a library for encoding the process of designing synthetic DNA
constructs. Coral mirrors the traditional design steps used in GUI-based
sequence design (ApE, j5, Benchling, etc.) as operations on data structures,
enables iterative design through analysis modules, and connects seamlessly to
outside libraries. Through the use of Coral, you can translate your DNA design
processes into concise, executable, and reusable scripts.

Coral encodes synthetic DNA design rules into its core sequence data types
(DNA, RNA, and Peptide), enabling concise, dependable methods for automated
DNA design.

Coral works with PyPy so long as a PyPy-compatible numpy is installed.
'''

# Check python versions
import sys
if sys.version_info.major > 2:
    print('Coral is currently compatible only with Python 2.')
    sys.exit(1)

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

import numpy  # noqa


doclines = __doc__.split('\n')

config = {
    'name': 'coral',
    'version': '0.4.3',
    'description': doclines[0],
    'long_description': '\n'.join(doclines[2:]),
    'author': 'Nick Bolten',
    'author_email': 'nbolten _at_ gmail',
    'maintainer': 'Nick Bolten',
    'maintainer_email': 'nbolten _at_ gmail',
    'url': 'https://github.com/klavinslab/coral',
    'license': 'MIT',
    'download_url': 'https://github.com/klavinslab/coral.git',
    'install_requires': ['numpy', 'biopython'],
    'extras_require': {'plotting': ['matplotlib'],
                       'yeastdatabases': ['intermine', 'requests'],
                       'documentation': ['sphinx']},
    'packages': ['coral',
                 'coral.analysis',
                 'coral.analysis._sequence',
                 'coral.analysis._sequencing',
                 'coral.analysis._structure',
                 'coral.constants',
                 'coral.database',
                 'coral.design',
                 'coral.design._oligo_synthesis',
                 'coral.design._sequence_generation',
                 'coral.seqio',
                 'coral.reaction',
                 'coral.sequence',
                 'coral.utils'],
    'package_data': {'coral': ['coral/analysis/_sequencing/data/*',
                               'coral/sequence/d3-plasmid.js']},
    'include_package_data': True,
    'scripts': [],
    'classifiers': ['Programming Language :: Python',
                    'Programming Language :: Python :: 2.7',
                    'Programming Language :: Python :: 2 :: Only',
                    'Topic :: Scientific/Engineering',
                    'Topic :: Scientific/Engineering :: Bio-Informatics']
}

seq_extension = Extension('coral.analysis._sequencing.calign',
                          ['coral/analysis/_sequencing/calign.c'],
                          include_dirs=[numpy.get_include()])
EXTENSIONS = [seq_extension]

setup(ext_modules=EXTENSIONS,
      test_suite='nose.collector',
      **config)



'''Gibson reaction simulation.'''
import logging
import warnings
import coral.analysis


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class AmbiguousGibsonError(ValueError):
    '''Exception to raise when Gibson is ambiguous.'''
    pass


class GibsonOverlapError(ValueError):
    '''Exception to raise when Gibson fails due to bad overlaps.'''
    pass


def gibson(seq_list, linear=False, homology=10, tm=63.0):
    '''Simulate a Gibson reaction.

    :param seq_list: list of DNA sequences to Gibson
    :type seq_list: list of coral.DNA
    :param linear: Attempt to produce linear, rather than circular,
                   fragment from input fragments.
    :type linear: bool
    :param homology_min: minimum bp of homology allowed
    :type homology_min: int
    :param tm: Minimum tm of overlaps
    :type tm: float
    :returns: coral.reaction.Gibson instance.
    :raises: ValueError if any input sequences are circular DNA.

    '''
    # FIXME: Preserve features in overlap
    # TODO: set a max length?
    # TODO: add 'expected' keyword argument somewhere to automate
    # validation

    # Remove any redundant (identical) sequences
    seq_list = list(set(seq_list))
    for seq in seq_list:
        if seq.circular:
            raise ValueError('Input sequences must be linear.')

    # Copy input list
    working_list = [s.copy() for s in seq_list]

    # Attempt to fuse fragments together until only one is left
    while len(working_list) > 1:
        working_list = _find_fuse_next(working_list, homology, tm)
    if not linear:
        # Fuse the final fragment to itself
        working_list = _fuse_last(working_list, homology, tm)

    # Clear features
    working_list[0].features = []
    return _annotate_features(working_list[0], seq_list)


def _annotate_features(template, fragment_list):
    locations = []
    template_copy = template.copy()
    for fragment in fragment_list:
        fwd_loc, rev_loc = template_copy.locate(fragment)
        for f in fwd_loc:
            start = f
            stop = f + len(fragment)
            locations.append((start, stop, fragment, 1))
        for r in rev_loc:
            start = len(template_copy) - r - len(fragment)
            stop = len(template_copy) - r
            locations.append((start, stop, fragment, -1))
    for start, stop, fragment, orient in sorted(locations, key=lambda x: x[0]):
        fragment = fragment.copy()
        if orient == -1:
            fragment = fragment.reverse_complement()
        for feature in fragment.features:
            feature_copy = feature.copy()
            feature_copy.move(start)
            template_copy.features.append(feature_copy)
    return template_copy


def _find_fuse_next(working_list, homology, tm):
    '''Find the next sequence to fuse, and fuse it (or raise exception).

    :param homology: length of terminal homology in bp
    :type homology: int
    :raises: AmbiguousGibsonError if there is more than one way for the
             fragment ends to combine.
             GibsonOverlapError if no homology match can be found.

    '''
    # 1. Take the first sequence and find all matches
    # Get graphs:
    #   a) pattern watson : targets watson
    #   b) pattern watson : targets crick
    #   c) pattern crick: targets watson
    #   d) pattern crick: targets crick
    pattern = working_list[0]
    targets = working_list[1:]

    # Output graph nodes of terminal binders:
    #   (destination, size, strand1, strand2)
    def graph_strands(strand1, strand2):
        graph = []
        for i, target in enumerate(targets):
            matchlen = homology_report(pattern, target, strand1, strand2,
                                       cutoff=homology, min_tm=tm)
            if matchlen:
                graph.append((i, matchlen, strand1, strand2))
        return graph

    graph_ww = graph_strands('w', 'w')
    graph_wc = graph_strands('w', 'c')
    graph_cw = graph_strands('c', 'w')
    graph_cc = graph_strands('c', 'c')
    graphs_w = graph_ww + graph_wc
    graphs_c = graph_cw + graph_cc
    graphs = graphs_w + graphs_c

    # 2. See if there's more than one result on a strand.
    # If so, throw an exception.
    if len(graphs_w) > 1 or len(graphs_c) > 1:
        raise AmbiguousGibsonError('multiple compatible ends.')
    if len(graphs_w) == len(graphs_c) == 0:
        raise GibsonOverlapError('Failed to find compatible Gibson ends.')

    # 3. There must be one result. Where is it?
    # If there's one result on each strand, go with the one that matches the
    # pattern watson strand (will occur first - index 0)
    match = graphs[0]

    # 4. Combine pieces together
    # 4a. Orient pattern sequence
    if match[2] == 'c':
        left_side = pattern.reverse_complement()
    else:
        left_side = pattern
    # 4b. Orient target sequence
    if match[3] == 'w':
        right_side = working_list.pop(match[0] + 1).reverse_complement()
    else:
        right_side = working_list.pop(match[0] + 1)

    working_list[0] = left_side + right_side[match[1]:]
    return working_list


def _fuse_last(working_list, homology, tm):
    '''With one sequence left, attempt to fuse it to itself.

    :param homology: length of terminal homology in bp.
    :type homology: int
    :raises: AmbiguousGibsonError if either of the termini are palindromic
             (would bind self-self).
             ValueError if the ends are not compatible.

    '''
    # 1. Construct graph on self-self
    #    (destination, size, strand1, strand2)
    pattern = working_list[0]

    def graph_strands(strand1, strand2):
        matchlen = homology_report(pattern, pattern, strand1, strand2,
                                   cutoff=homology, min_tm=tm, top_two=True)
        if matchlen:
            # Ignore full-sequence matches
            # HACK: modified homology_report to accept top_two. It should
            # really just ignore full-length matches
            for length in matchlen:
                if length != len(pattern):
                    return (0, length, strand1, strand2)
        else:
            return []

    # cw is redundant with wc
    graph_ww = graph_strands('w', 'w')
    graph_wc = graph_strands('w', 'c')
    graph_cc = graph_strands('c', 'c')
    if graph_ww + graph_cc:
        raise AmbiguousGibsonError('Self-self binding during circularization.')
    if not graph_wc:
        raise ValueError('Failed to find compatible ends for circularization.')

    working_list[0] = working_list[0][:-graph_wc[1]].circularize()

    return working_list


def homology_report(seq1, seq2, strand1, strand2, cutoff=0, min_tm=63.0,
                    top_two=False, max_size=500):
    '''Given two sequences (seq1 and seq2), report the size of all perfect
    matches between the 3' end of the top strand of seq1 and the 3' end of
    either strand of seq2. In short, in a Gibson reaction, what would bind the
    desired part of seq1, given a seq2?

    :param seq1: Sequence for which to test 3\' binding of a single strand to
                 seq2.
    :type seq1: coral.DNA
    :param seq2: Sequence for which to test 3\' binding of each strand to seq1.
    :type seq1: coral.DNA
    :param strand1: w (watson) or c (crick) - which strand of seq1 is being
                        tested.
    :type strand1ed: str
    :param strand2: w (watson) or c (crick) - which strand of seq2 is being
                        tested.
    :type strand2ed: str
    :param cutoff: size cutoff for the report - if a match is lower, it's
                   ignored
    :type cutoff: int
    :param min_tm: Minimum tm value cutoff - matches below are ignored.
    :type min_tm: float
    :param top_two: Return the best two matches
    :type top_two: bool
    :param max_size: Maximum overlap size (increases speed)
    :type max_size: int
    :returns: List of left and right identities.
    :rtype: list of ints

    '''
    # Ensure that strand 1 is Watson and strand 2 is Crick
    if strand1 == 'c':
        seq1 = seq1.reverse_complement()
    if strand2 == 'w':
        seq2 = seq2.reverse_complement()
    # Generate all same-length 5' ends of seq1 and 3' ends of seq2 within
    # maximum homology length
    # TODO: If strings aren't used here, gen_chunks takes forever. Suggests a
    # need to optimize coral.DNA subsetting
    seq1_str = str(seq1)
    seq2_str = str(seq2)

    def gen_chunks(s1, s2):
        chunks1 = [seq1_str[-(i + 1):] for i in range(min(len(seq1_str),
                                                      max_size))]
        chunks2 = [seq2_str[:(i + 1)] for i in range(min(len(seq2_str),
                                                     max_size))]
        return chunks1, chunks2

    seq1_chunks, seq2_chunks = gen_chunks(seq1_str, seq2_str)

    # Check for exact matches from terminal end to terminal end
    target_matches = []
    for i, (s1, s2) in enumerate(zip(seq1_chunks, seq2_chunks)):
        s1len = len(s1)
        # Inefficient! (reverse complementing a bunch of times)
        # Don't calculate tm once base tm has been reached.
        # TODO: Go through logic here again and make sure the order of checking
        # makes sense
        if s1 == s2:
            logger.debug('Found Match: {}'.format(str(s1)))
            if s1len >= cutoff:
                tm = coral.analysis.tm(seq1[-(i + 1):])
                logger.debug('Match tm: {} C'.format(tm))
                if tm >= min_tm:
                    target_matches.append(s1len)
                elif tm >= min_tm - 4:
                    msg = 'One overlap had a Tm of {} C.'.format(tm)
                    warnings.warn(msg)
                    target_matches.append(s1len)

    target_matches.sort()
    if not top_two:
        return 0 if not target_matches else target_matches[0]
    else:
        return 0 if not target_matches else target_matches[0:2]
